# -*- coding: utf-8 -*-

"""
DSuper8 project based on Joe Herman's rpi-film-capture.

Software modified by Manuel Ángel.

User interface redesigned by Manuel Ángel.

DS8Dialogs.py: Module for the visualization of the captured images, calculation
               and visualization of the histogram and management of the
               graphical user interface.

Last version: July 2021.
"""

from PyQt5.QtGui import QImage, QPaintEvent, QPainter, QIcon, QColor

from PyQt5.QtCore import Qt, pyqtSlot

from PyQt5.QtWidgets import QDialog, QFileDialog, QMessageBox, QApplication

from threading import Event

from numpy import array, zeros, float32, ndarray

from cv2 import split, calcHist, cvtColor, COLOR_BGR2RGB, resize

import matplotlib.pyplot as plt

from time import sleep

from pathlib import Path

from glob import glob

from logging import info

from configparser import Error

# Our own modules.

import config

import DS8Config

from codes import *

# Main dialog class, created with Qt Designer and translated into Python using
# pyuic5.
from ui_DSuper8 import *

# This class is used to display the images sent by the server.


class DS8ImgDialog(QDialog):
    def __init__(self, parent=None):
        super(DS8ImgDialog, self).__init__(parent)
        self.icon = QIcon(":/DSuper8Icon.png")
        self.setWindowIcon(self.icon)
        # The close button is removed from the window.
        self.setWindowFlags(Qt.Window |
                            Qt.WindowTitleHint |
                            Qt.WindowMinimizeButtonHint |
                            Qt.WindowMaximizeButtonHint)
        self.move(10, 10)
        self.setWindowTitle(config.imgWinTitle)
        self.mQImage = None
        self.displayImgEvent = Event()

    def displayImg(self, cvimg, title):
        #info("Viewing image")
        self.setWindowTitle("Image: " + title)
        self.cvImage = cvtColor(cvimg, COLOR_BGR2RGB)
        h, w, byteValue = self.cvImage.shape
        config.imgCapResH = h
        config.imgCapResW = w
        byteValue = byteValue * w
        imgwinh = config.imgWinHeight
        imgwinw = int((imgwinh / h) * w)
        if imgwinw > config.imgWinWidth:
            imgwinw = config.imgWinWidth
            imgwinh = int((imgwinw / w) * h)

        self.resize(imgwinw, imgwinh)
        self.mQImage = QImage(self.cvImage, w, h, byteValue,
                              QImage.Format_RGB888)
        self.update()

    def paintEvent(self, QPaintEvent):
        #info("paintEvent called")
        painter = QPainter()
        painter.begin(self)
        if self.mQImage:
            painter.drawImage(0, 0, QImage.scaled(self.mQImage, self.width(),
                                                  self.height(),
                                                  Qt.KeepAspectRatio))
        painter.end()
        self.displayImgEvent.set()

    def setupThreadingUpdates(self, imgthread):
        info("Setting up thread updates for the image window")
        imgthread.displayImgSig.connect(self.displayImg)
        self.displayImgEvent = imgthread.displayImgEvent


# This class is used to display the histograms corresponding to the images
# sent by the server.
class DS8Histogram():
    def __init__(self):

        self.xnpa = array(range(256))
        self.b_ynpa = zeros(256, dtype=float32)
        self.g_ynpa = zeros(256, dtype=float32)
        self.r_ynpa = zeros(256, dtype=float32)

        plt.rcParams["toolbar"] = "None"
        self.fig = plt.figure(num=None, figsize=(4.8, 3.6), dpi=100,
                              facecolor="w", edgecolor="k")
        self.icon = QIcon(":/DSuper8Icon.png")
        self.fig.canvas.parent().setWindowIcon(self.icon)
        # The close button is removed from the window.
        self.fig.canvas.parent().setWindowFlags(Qt.Window |
                                                Qt.WindowTitleHint |
                                                Qt.WindowMinimizeButtonHint |
                                                Qt.WindowMaximizeButtonHint)
        self.fig.canvas.set_window_title("Histogram")
        self.mngr = plt.get_current_fig_manager()

        self.plot = plt.subplot(111)
        plt.ion()
        plt.show()

    def plotHistogram(self, cvimg, title):
        #info("plotHistogram called")

        self.fig.canvas.set_window_title("Histogram: " + title)

        bgr_planes = split(cvimg)

        histSize = 256
        histRange = (0, 256)
        accumulate = False

        b_hist = calcHist(bgr_planes, [0], None, [histSize], histRange,
                          accumulate=accumulate)
        g_hist = calcHist(bgr_planes, [1], None, [histSize], histRange,
                          accumulate=accumulate)
        r_hist = calcHist(bgr_planes, [2], None, [histSize], histRange,
                          accumulate=accumulate)

        for i in range(256):
            self.b_ynpa[i] = b_hist[i][0]
            self.g_ynpa[i] = g_hist[i][0]
            self.r_ynpa[i] = r_hist[i][0]

        # Delete old chart.
        self.plot.cla()

        # Set x-axis limits.
        self.plot.set_xlim([0, 255])

        # Set logarithmic scale on y-axis.
        if config.logarithmHist:
            plt.yscale("log", nonpositive="clip")

        # Make the new chart.
        plt.plot(self.xnpa, self.b_ynpa, linewidth=1, color="blue")
        plt.plot(self.xnpa, self.g_ynpa, linewidth=1, color="green")
        plt.plot(self.xnpa, self.r_ynpa, linewidth=1, color="red")

        self.fig.canvas.draw_idle()

    # Close the histogram window.
    def closeHist(self):
        #info("Closing the histogram called")
        plt.close(self.fig)

# Main dialog class for displaying and managing the user interface, created
# with Qt Designer.


class DS8Dialog(QDialog, Ui_DSuper8):
    def __init__(self, parent=None):
        super(DS8Dialog, self).__init__(parent)
        self.setupUi(self)
        self.icon = QIcon(":/DSuper8Icon.png")
        self.setWindowIcon(self.icon)
        self.move(config.monitor_res - 10 - self.width(), 10)
        self.config = DS8Config.DS8ConfigParser()
        self.configFile = Path(config.configfile)
        self.histograma = DS8Histogram()
        self.plotHistogramEvent = Event()

        # Geometry of the region of interest to be captured by the camera.

        # Image size captured by camera.
        self.imgCapW = 2028
        self.imgCapH = 1520

        # Coordinates of the upper left corner.
        self.roiX = 0
        self.roiY = 0

        # Zoom value. Determine the width and height of the region of interest.
        # Values between 400 and 1000. They are given by the zoom control.
        self.roiZ = 1000

        # Scrolling the region of interest each time a button is pressed.
        self.deltaRoi = 0

        # Blue and red gain for manual white balance adjustment.
        if config.cameraType == "V1":
            self.manualBlueGain = 100
            self.manualRedGain = 100
        elif config.cameraType == "HQ":
            self.manualBlueGain = 160
            self.manualRedGain = 320

        # Manual exposure time.
        self.manExpTime = 0.0

        # Denotes if the motor is in motion or at rest.
        # Repose -> True - Movement -> False
        self.motorNotMoving = True

        # Indicates that the file naming index is out of sync with the position
        # of the movie.
        self.dirty = True

        # Last image viewed.
        # Initializes with Splash image.
        self.lastShowImg = config.readImgFromQrc(":/DSuper8Splash.jpg")

        # Title of the last image viewed.
        self.lastShowTitle = "Splash"
        
        # Capture folder is determined.
        self.setCaptureFolder()

    # By default, the user interface is designed to use the HQ camera.
    # This function is to adapt the UI to use the Raspberry Pi V1 camera
    def V1CamGUI(self):
        # Title.
        self.groupBox_2.setTitle("V1 camera settings")
        
        # Modification of blue and red gains settings.
        self.blueLabel.setToolTip("Blue gain for manual white balance\n"
                                  "adjustment.\n"
                                  "Values between 50 and 200.\n"
                                  "By default 100.")
        self.blueGain.setToolTip("Blue gain for manual white balance\n"
                                 "adjustment.\n"
                                 "Values between 50 and 200.\n"
                                 "By default 100.")
        self.blueGain.setMaximum(200)
        self.blueGain.setValue(128)
        self.blueGainBox.setToolTip("Blue gain for manual white balance\n"
                                    "adjustment.\n"
                                    "Values between 50 and 200.\n"
                                    "By default 100.")
        self.blueGainBox.setMaximum(200)
        self.blueGainBox.setValue(128)
        self.blueResetBtn.setToolTip("Initial value of blue at 100")
        self.redLabel.setToolTip("Red gain for manual white balance\n"
                                 "adjustment.\n"
                                 "Values between 50 and 200.\n"
                                 "By default 100.")
        self.redGain.setToolTip("Red gain for manual white balance\n"
                                "adjustment.\n"
                                "Values between 50 and 200.\n"
                                "By default 100.")
        self.redGain.setMaximum(200)
        self.redGain.setValue(137)
        self.redGainBox.setToolTip("Red gain for manual white balance\n"
                                   "adjustment.\n"
                                   "Values between 50 and 200.\n"
                                   "By default 100.")
        self.redGainBox.setMaximum(200)
        self.redGainBox.setValue(137)
        self.redResetBtn.setToolTip("Initial value of red at 100")
       
        # Modification of resolutions.
        self.resolutionBox.setItemText(0, "2592x1944 px")
        self.resolutionBox.setItemText(1, "1600x1200 px")
        self.resolutionBox.setItemText(2, "1440x1080 px")
        self.resolutionBox.addItem("1296x972 px")
        self.resolutionBox.addItem("1024x768 px")
        self.resolutionBox.addItem("800x600 px")

    # This function is used to send control sequences to the server.
    def sendCtrl(self, cmd):
        info("CTRL: " + cmd)
        config.ctrl_conn.write(cmd + "\n")
        config.ctrl_conn.flush()

    # Configuring connections to UI update from imgthread.
    def setupThreadingUpdates(self, imgthread):
        imgthread.updateFrameNumSig.connect(self.updateFrameNum)
        imgthread.updateAESig.connect(self.updateSSAE)
        imgthread.updateSSSig.connect(self.updateSS)
        imgthread.updateGainsSig.connect(self.updateGains)
        imgthread.updateStatusSig.connect(self.updateStatus)
        imgthread.displayImgSig.connect(self.takenImg)
        imgthread.plotHistogramSig.connect(self.updateHistogram)
        imgthread.motorStoppedSig.connect(self.motorStopped)
        imgthread.lightSig.connect(self.updateLightCheckbox)
        imgthread.frameLimitReachedSig.connect(self.captureEnd)
        self.plotHistogramEvent = imgthread.plotHistogramEvent

    # This function is used to send the UI settings to the server.
    # Sent at startup.
    def sendInitConfig(self):

        # Zoom dial setting.
        self.sendCtrl(SETZ + str(self.zoomDial.value()))

        # Setting the region of interest.
        self.sendCtrl(SETX + str(self.roiX))
        self.sendCtrl(SETY + str(self.roiY))

        # Lighting control.
        if self.lightCheckbox.isChecked():
            self.sendCtrl(LIGHT_ON)
        else:
            self.sendCtrl(LIGHT_OFF)

        # Sensitivity adjustment.
        iso = self.ISOcomboBox.currentIndex()
        self.setISO(iso)

        # Exposure compensation.
        self.sendCtrl(EXP_COMP + str(self.EV.value()))

        # White balance.
        awb = self.awbBox.currentIndex()
        self.setAWBmode(awb)

        # Brightness adjustment.
        self.sendCtrl(BRIGHTNESS + str(self.brightness.value()))
        
        # Contrast adjustment.
        self.sendCtrl(CONTRAST + str(self.contrastBox.value()))

        # Saturation adjustment.
        self.sendCtrl(SATURATION + str(self.saturationBox.value()))

        # Adjustment bracketing exposures.
        bracketing = self.bracketingBox.value()
        if bracketing > 1:
            config.exposureTimes.resize(bracketing, refcheck=False)
        self.sendCtrl(BRACKETING_SHOTS + str(bracketing))
        stops = self.stopsBox.value()
        self.sendCtrl(BRACKETING_STOPS + str(stops))

        # Auto exposure control.
        if self.autoExpCheckBox.isChecked():
            self.sendCtrl(AUTOEXP_ON)
        else:
            # Manual exposure time adjustment.
            self.sendCtrl(AUTOEXP_OFF + str(self.timeExpBox.value()))

        # Vertical and horizontal flip.
        if self.vflip.isChecked():
            self.sendCtrl(VFLIP_ON)
        else:
            self.sendCtrl(VFLIP_OFF)

        if self.hflip.isChecked():
            self.sendCtrl(HFLIP_ON)
        else:
            self.sendCtrl(HFLIP_OFF)

        # Black and white image.
        if self.bw.isChecked():
            self.sendCtrl(BW_ON)
        else:
            self.sendCtrl(BW_OFF)

        # DRC selection.
        self.setDrc(self.drcBox.currentIndex())

        # Sharpness adjustment.
        self.sendCtrl(SHARPNESS + str(self.sharpnessBox.value()))

        # Capture resolution.
        self.sendCtrl(SET_SIZE + str(self.resolutionBox.currentIndex()))

        # Still Image Port.
        if self.stillCap.isChecked():
            self.sendCtrl(VID_PORT_OFF)
        else:
            self.sendCtrl(VID_PORT_ON)

    # Capture folder is determined.
    def setCaptureFolder(self):
        # If there is a valid folder in the .conf file, it is selected.
        capFolder = self.capFolderBox.text().strip()
        
        if capFolder and Path(capFolder).is_dir():
            
            return      

        # If it is not valid, it is tried with the folder indicated in the
        # global variables file config.py.

        elif (config.capFolder.strip() 
              and Path(config.capFolder.strip()).is_dir()):
            self.capFolderBox.setText(config.capFolder.strip())
            if self.startPosCheckBox.isChecked():
                self.captureTestBtn.setEnabled(True)
                self.captureStartBtn.setEnabled(True)

        # Finally, if it is not valid, it is left blank and the user must
        # manually select the folder before starting to scan.

        else:
            self.capFolderBox.setText("Select capture folder")
            config.capFolder = ""
            self.captureTestBtn.setEnabled(False)
            self.captureStartBtn.setEnabled(False)

    # Post-capture settings.
    def initPostCapture(self):

        # Show histogram.
        if self.showHist.isChecked():
            self.initHistogram(True)

        else:
            self.initHistogram(False)

        # Y axis of the histogram logarithmic.
        if self.logarithmHist.isChecked():
            self.mklogHist(True)
        else:
            self.mklogHist(False)

        # Round corners.
        if self.roundCorns.isChecked():
            self.setRoundCorns(True)
        else:
            self.setRoundCorns(False)

        # Rotate image.
        if self.rotation.isChecked():
            self.setRotation(True)
        else:
            self.setRotation(False)

        # Crop image edges.
        if self.cropping.isChecked():
            self.setCrop(True)
        else:
            self.setCrop(False)
            
        # Enable HDR tab.
        if self.bracketingBox.value() == 1:        
            self.stopsBox.setEnabled(False)
            self.DS8tabWidget.setTabEnabled(4, False)
            self.setHDRAlgorithm(False)
            self.setToneMapAlgorithm(False)
            self.stopsBox.setValue(0.0)

        else:
            config.exposureTimes.resize(self.bracketingBox.value(),
                                        refcheck=False)
            self.stopsBox.setEnabled(True)
            self.DS8tabWidget.setTabEnabled(4, True)
            self.setHDRAlgorithm(False)
            self.setToneMapAlgorithm(False)

    # Slots connected to the signals generated by the UI.

    # Start.

    # zoomDial

    def setZoom(self, zoomval):
        self.roiZ = zoomval
        self.roiX = int((1000 - self.roiZ) / 2)
        self.roiY = self.roiX
        self.deltaRoi = int((1000 - self.roiZ) / 10)
        if self.roiZ == 1000:
            self.roiUpButton.setEnabled(False)
            self.roiLeftButton.setEnabled(False)
            self.roiRightButton.setEnabled(False)
            self.roiDownButton.setEnabled(False)
        elif self.roiZ != 1000 and self.prevCheckBox.isChecked():
            self.roiUpButton.setEnabled(True)
            self.roiLeftButton.setEnabled(True)
            self.roiRightButton.setEnabled(True)
            self.roiDownButton.setEnabled(True)

        self.sendCtrl(SETZ + str(self.roiZ))

        # Capture size and maximum allowed cropping are updated.
        self.setMaxCrop()

        self.updateStatus("Zoom level set")

    # roiUpButton
    def setRoiYUp(self):
        if self.roiY - self.deltaRoi >= 0:
            self.roiY -= self.deltaRoi
        else:
            self.roiUpButton.setEnabled(False)

        if self.roiY < self.deltaRoi:
            self.roiUpButton.setEnabled(False)

        if self.roiY + self.roiZ + self.deltaRoi <= 1000:
            self.roiDownButton.setEnabled(True)

        self.sendCtrl(SETY + str(self.roiY))
        self.updateStatus("ROI shifted up")

    # roiLeftButton
    def setRoiXLeft(self):
        if self.roiX - self.deltaRoi >= 0:
            self.roiX -= self.deltaRoi
        else:
            self.roiLeftButton.setEnabled(False)

        if self.roiY < self.deltaRoi:
            self.roiUpButton.setEnabled(False)

        if self.roiX + self.roiZ + self.deltaRoi < 1000:
            self.roiRightButton.setEnabled(True)

        self.sendCtrl(SETX + str(self.roiX))
        self.updateStatus("ROI shifted to the left")

    # roiRightButton
    def setRoiXRight(self):
        if self.roiX + self.roiZ + self.deltaRoi < 1000:
            self.roiX += self.deltaRoi
        else:
            self.roiRightButton.setEnabled(False)

        if self.roiX > 1000 - self.deltaRoi:
            self.roiRightButton.setEnabled(False)

        if self.roiX - self.deltaRoi > 0:
            self.roiLeftButton.setEnabled(True)

        self.sendCtrl(SETX + str(self.roiX))
        self.updateStatus("ROI shifted to the right")

    # roiDownButton
    def setRoiYDown(self):
        if self.roiY + self.roiZ + self.deltaRoi <= 1000:
            self.roiY += self.deltaRoi
        else:
            self.roiDownButton.setEnabled(False)

        if self.roiY > 1000 - self.deltaRoi:
            self.roiDownButton.setEnabled(False)

        if self.roiY - self.deltaRoi >= 0:
            self.roiUpButton.setEnabled(True)

        self.sendCtrl(SETY + str(self.roiY))
        self.updateStatus("ROI shifted down")

    # lightCheckbox
    def lightSet(self, isOn):
        self.sendCtrl(LIGHT_ON if isOn else LIGHT_OFF)
        if isOn:
            self.updateStatus("Light on")
        else:
            self.updateStatus("Light off")

    # prevCheckBox
    def previewSet(self, isOn):
        if isOn:
            config.lastMode = "P"
            config.prevOn = True
            config.testImg = False
            config.captureOn = False
            self.sendCtrl(PREVIEW_ON)

            # Enable preview controls.
            self.zoomDial.setEnabled(True)
            if self.zoomDial.value() != 1000:
                self.roiUpButton.setEnabled(True)
                self.roiLeftButton.setEnabled(True)
                self.roiRightButton.setEnabled(True)
                self.roiDownButton.setEnabled(True)
            self.vflip.setEnabled(True)
            self.hflip.setEnabled(True)
            self.bw.setEnabled(True)
            self.drcLabel.setEnabled(True)
            self.drcBox.setEnabled(True)
            self.sharpnessLabel.setEnabled(True)
            self.sharpnessBox.setEnabled(True)
            self.resolutionLabel.setEnabled(True)
            self.resolutionBox.setEnabled(True)
            self.stillCap.setEnabled(True)
            self.roundCorns.setEnabled(True)
            self.rotation.setEnabled(True)
            if self.rotation.isChecked():
                self.rotationBox.setEnabled(True)
            self.cropping.setEnabled(True)
            if self.cropping.isChecked():
                self.cropTopLabel.setEnabled(True)
                self.cropTopBox.setEnabled(True)
                self.cropLeftLabel.setEnabled(True)
                self.cropLeftBox.setEnabled(True)
                self.cropRightLabel.setEnabled(True)
                self.cropRightBox.setEnabled(True)
                self.cropBottomLabel.setEnabled(True)
                self.cropBottomBox.setEnabled(True)

            # Disable controls.
            self.captureTestBtn.setEnabled(False)
            self.captureStartBtn.setEnabled(False)
            self.capturePauseBtn.setEnabled(False)
            self.loadConfigButton.setEnabled(False)

            # Disable exit button.
            self.quitButton.setEnabled(False)

            # Request image to server avoiding imgThread crash.
            self.sendCtrl(NEW_IMAGE)

            info("Waiting for image " + str(config.numImgRec))
            self.updateStatus("Preview images")

        else:
            config.prevOn = False
            self.sendCtrl(PREVIEW_OFF)
            self.loadConfigButton.setEnabled(True)

            # Disable preview controls.
            self.zoomDial.setEnabled(False)
            self.roiUpButton.setEnabled(False)
            self.roiLeftButton.setEnabled(False)
            self.roiRightButton.setEnabled(False)
            self.roiDownButton.setEnabled(False)
            self.vflip.setEnabled(False)
            self.hflip.setEnabled(False)
            self.bw.setEnabled(False)
            self.drcLabel.setEnabled(False)
            self.drcBox.setEnabled(False)
            self.sharpnessLabel.setEnabled(False)
            self.sharpnessBox.setEnabled(False)
            self.roundCorns.setEnabled(False)
            self.rotation.setEnabled(False)
            self.rotationBox.setEnabled(False)
            self.cropping.setEnabled(False)
            self.cropTopLabel.setEnabled(False)
            self.cropTopBox.setEnabled(False)
            self.cropLeftLabel.setEnabled(False)
            self.cropLeftBox.setEnabled(False)
            self.cropRightLabel.setEnabled(False)
            self.cropRightBox.setEnabled(False)
            self.cropBottomLabel.setEnabled(False)
            self.cropBottomBox.setEnabled(False)

            # Enable controls.
            if self.motorNotMoving:
                if Path(config.capFolder.strip()).is_dir():
                    enable = True
                else:
                    enable = False
                if self.startPosCheckBox.isChecked():
                    self.captureTestBtn.setEnabled(enable)
                    self.captureStartBtn.setEnabled(enable)
                self.capturePauseBtn.setEnabled(False)

                # Enable exit button.
                self.quitButton.setEnabled(True)

    # startPosCheckBox
    def synchronizePositionIndicator(self, checked):

        filmPosition = int(self.frameLcd.value())

        if self.startPosCheckBox.isChecked():

            # We tell the server to send position update signals.
            self.sendCtrl(UPDATE_FRAME)

            if filmPosition:
                config.file_number = filmPosition
                config.frame_number = filmPosition
            else:
                config.file_number = 1
                config.frame_number = 1
                self.frameLcd.display(1)

            self.dirty = False
            self.frameLcd.setStyleSheet("""QLCDNumber {background-color:
                                        rgb(245, 255, 172)}""")
            self.frameLcd.setEnabled(True)
            self.gotoCheckbox.setEnabled(True)
            self.passToCheckbox.setEnabled(True)
            self.nextFrameBox.setEnabled(True)
            self.frameLCDCountUpBtn.setEnabled(True)
            self.frameLCDCountDownBtn.setEnabled(True)
            self.captureFrmRev10.setEnabled(True)
            self.captureFrmRev.setEnabled(True)
            self.captureFrmAdv.setEnabled(True)
            self.captureFrmAdv10.setEnabled(True)
            if (config.capFolder.strip()
                and Path(config.capFolder.strip()).is_dir()
                and not config.prevOn):
                self.captureTestBtn.setEnabled(True)
                self.captureStartBtn.setEnabled(True)

            filmPosition = int(self.frameLcd.value())

            if filmPosition == 1:
                self.updateStatus("Film in initial position")

            elif filmPosition > 1:
                self.updateStatus("Set position of the film")
        else:

            # We tell the server not to send position update signals.
            self.sendCtrl(NO_UPDATE_FRAME)

            config.file_number = 0
            config.frame_number = 0
            self.dirty = True
            self.captureTestBtn.setEnabled(False)
            self.frameLcd.setStyleSheet("""QLCDNumber {background-color:
                                        rgb(241, 241, 236)}""")
            self.frameLcd.setEnabled(False)
            self.frameLcd.display(0)
            self.gotoCheckbox.setEnabled(False)
            self.passToCheckbox.setEnabled(False)
            self.nextFrameBox.setEnabled(False)
            self.frameLCDCountUpBtn.setEnabled(False)
            self.frameLCDCountDownBtn.setEnabled(False)
            self.captureFrmRev10.setEnabled(False)
            self.captureFrmRev.setEnabled(False)
            self.captureFrmAdv.setEnabled(False)
            self.captureFrmAdv10.setEnabled(False)
            self.captureStartBtn.setEnabled(False)
            self.updateStatus("Film is not in initial position")

    # fRevButton
    def motorfrev(self):
        # Motor is activated.
        if not self.activateMotorCheckBox.isChecked():
            self.sendCtrl(ACTIVATE_MOTOR)
            self.activateMotorCheckBox.setChecked(True)
        self.sendCtrl(MOTOR_FREV)
        if self.startPosCheckBox.isChecked():
            self.dirty = True
        self.motorNotMoving = False
        self.updateStatus("Continuous reverse motor")
        self.stopButton.setEnabled(True)
        self.disableButtons()

    # reverseButton
    def motorRev(self):
        # Motor is activated.
        if not self.activateMotorCheckBox.isChecked():
            self.sendCtrl(ACTIVATE_MOTOR)
            self.activateMotorCheckBox.setChecked(True)
        self.sendCtrl(MOTOR_REV)
        if self.startPosCheckBox.isChecked():
            self.dirty = True
        self.motorNotMoving = False
        self.updateStatus("1 frame reverse")
        self.stopButton.setEnabled(False)
        self.disableButtons()

    # stopButton
    def motorStop(self):
        self.sendCtrl(MOTOR_STOP)
        self.updateStatus("Engine stop")
        self.stopButton.setEnabled(False)
        # The self.enableButtons() function is called when the motorStoppedSig
        # signal is received.

    # forwardButton
    def motorFwd(self):
        # Motor is activated.
        if not self.activateMotorCheckBox.isChecked():
            self.sendCtrl(ACTIVATE_MOTOR)
            self.activateMotorCheckBox.setChecked(True)
        self.sendCtrl(MOTOR_FWD)
        if self.startPosCheckBox.isChecked():
            self.dirty = True
        self.motorNotMoving = False
        self.updateStatus("1 frame advance")
        self.stopButton.setEnabled(False)
        self.disableButtons()

    # ffdButton
    def motorffd(self):
        # Motor is activated.
        if not self.activateMotorCheckBox.isChecked():
            self.sendCtrl(ACTIVATE_MOTOR)
            self.activateMotorCheckBox.setChecked(True)
        self.sendCtrl(MOTOR_FFD)
        if self.startPosCheckBox.isChecked():
            self.dirty = True
        self.motorNotMoving = False
        self.updateStatus("Continuous advance motor")
        self.stopButton.setEnabled(True)
        self.disableButtons()

    # configFileBox
    def checkconfigFileBox(self, text):
        # Text is checked.
        if not self.configFileBox.text():
            self.saveConfigButton.setEnabled(False)
        else:
            self.saveConfigButton.setEnabled(True)

    # loadConfigButton
    def loadConfig(self):
        # Self.config is restarted.
        for section in self.config.sections():
            self.config.remove_section(section)

        # Configuration file selection dialog.
        filtros = ["*.conf"]
        fileDialog = QFileDialog()
        fileDialog.setFileMode(QFileDialog.ExistingFiles)
        fileDialog.setWindowTitle("Select configuration file")
        fileDialog.setDirectory(config.default_folder)
        fileDialog.setNameFilters(filtros)

        if fileDialog.exec():

            self.configFile = Path(fileDialog.selectedFiles()[0])

        else:
            return

        if self.configFile.is_file():
            configError = False
            info("Reading " + str(self.configFile) + " configuration file")

            try:
                self.config.read([self.configFile])
                if not self.config.validateConfigFile():
                    raise Error

            except Error:
                # Pop-up message if there are errors loading the configuration
                # file.
                msgBox = QMessageBox()
                msgBox.setWindowTitle("Invalid configuration file")
                msgBox.setIcon(QMessageBox.Warning)
                msgBox.setText(str(self.configFile)
                               + " is not a valid configuration file")
                acceptButton = msgBox.addButton("Accept",
                                                QMessageBox.AcceptRole)

                msgBox.exec()

                configError = True
                info(str(self.configFile) + " is not a valid configuration file")

        else:
            configError = True
            info("Configuration file " + str(self.configFile)
                 + " does not exist")

        if not configError:
            self.config.updateUIfromConfig(self)
            self.configFileBox.setText(str(self.configFile))
            info(str(self.configFile) + " configuration file loaded")
                 
            self.updateStatus("Configuration file loaded")

        # Setting the region of interest.
        # You have to send the configuration because it does not have a slot.
        self.sendCtrl(SETX + str(self.roiX))
        self.sendCtrl(SETY + str(self.roiY))

        # Maximum allowed clipping is updated.
        self.setMaxCrop()

    # saveConfigButton
    def saveCurrentConfig(self):
        if self.configFile.is_file():
            self.config.updateConfigFromUI(self)
            with open(self.configFile, "w") as savefile:
                self.config.write(savefile)
            self.updateStatus("Configuration file saved")

        else:
            self.saveConfig()

    # saveAsButton
    def saveConfig(self):

        filtros = ["*.conf"]
        fileDialog = QFileDialog()
        fileDialog.setFileMode(QFileDialog.AnyFile)        
        fileDialog.setWindowTitle("Configuration file - Save as")
        fileDialog.setDirectory(config.default_folder)
        fileDialog.setNameFilters(filtros)

        if fileDialog.exec():
            self.configFile = fileDialog.selectedFiles()[0]

            if not self.configFile.endswith(".conf"):
                self.configFile += ".conf"

            self.configFile = Path(self.configFile)

        else:
            return

        self.config.updateConfigFromUI(self)
        with open(self.configFile, "w") as savefile:
            self.config.write(savefile)
        self.configFileBox.setText(str(self.configFile))
        self.saveConfigButton.setEnabled(True)
        self.updateStatus("Configuration file saved")

    # Camera settings.    

    # ISOcomboBox
    def setISO(self, iso):
        if iso == 0:
            self.sendCtrl(ISO + "0")
        elif iso == 1:
            self.sendCtrl(ISO + "100")
        elif iso == 2:
            self.sendCtrl(ISO + "200")
        elif iso == 3:
            self.sendCtrl(ISO + "320")
        elif iso == 4:
            self.sendCtrl(ISO + "400")
        elif iso == 5:
            self.sendCtrl(ISO + "500")
        elif iso == 6:
            self.sendCtrl(ISO + "640")
        elif iso == 7:
            self.sendCtrl(ISO + "800")

        self.updateStatus("Set ISO sensitivity")

    # EV
    def setEV(self, ev):
        self.sendCtrl(EXP_COMP + str(ev))
        self.updateStatus("Set exposure value")

    # awbBox
    def setAWBmode(self, awb):
        if config.cameraType == "V1":
            self.setAWBmodeV1(awb)

        elif config.cameraType == "HQ":
            self.setAWBmodeHQ(awb)

    # White balance for V1 camera.
    # awb_mode = "off" allows the camera's white balance to be set manually, by
    # adjusting the awb_gains parameter.
    # It was preferred to configure the gains manually. In tests carried out,
    # it has been verified that the property awb_mode does not work correctly
    # with the V1 camera.
    def setAWBmodeV1(self, awb):
        self.redGain.blockSignals(True)
        self.redGainBox.blockSignals(True)
        self.blueGain.blockSignals(True)
        self.blueGainBox.blockSignals(True)
        if awb == 0:
            mode = "off"
            self.blueGain.setSliderPosition(self.manualBlueGain)
            self.blueGainBox.setValue(self.manualBlueGain)
            self.redGain.setSliderPosition(self.manualRedGain)
            self.redGainBox.setValue(self.manualRedGain)
            self.sendCtrl(AWB_MODE + mode)
            self.sendCtrl(GAIN_BLUE + str(self.blueGainBox.value()))
            self.sendCtrl(GAIN_RED + str(self.redGainBox.value()))
        # Gain values for 3300 K color temperature.
        elif awb == 1:
            self.sendCtrl(AWB_MODE + "off")
            self.sendCtrl(GAIN_BLUE + str(128))
            self.sendCtrl(GAIN_RED + str(137))
            self.blueGain.setSliderPosition(128)
            self.blueGainBox.setValue(128)
            self.redGain.setSliderPosition(137)
            self.redGainBox.setValue(137)
        elif awb == 2:
            mode = "auto"
            self.sendCtrl(AWB_MODE + mode)
        elif awb == 3:
            mode = "sunlight"
            self.sendCtrl(AWB_MODE + "off")
            self.sendCtrl(GAIN_BLUE + str(121))
            self.sendCtrl(GAIN_RED + str(150))
            self.blueGain.setSliderPosition(121)
            self.blueGainBox.setValue(121)
            self.redGain.setSliderPosition(150)
            self.redGainBox.setValue(150)
        elif awb == 4:
            mode = "cloudy"
            self.sendCtrl(AWB_MODE + "off")
            self.sendCtrl(GAIN_BLUE + str(112))
            self.sendCtrl(GAIN_RED + str(158))
            self.blueGain.setSliderPosition(112)
            self.blueGainBox.setValue(112)
            self.redGain.setSliderPosition(158)
            self.redGainBox.setValue(158)
        elif awb == 5:
            mode = "shade"
            self.sendCtrl(AWB_MODE + "off")
            self.sendCtrl(GAIN_BLUE + str(112))
            self.sendCtrl(GAIN_RED + str(158))
            self.blueGain.setSliderPosition(112)
            self.blueGainBox.setValue(112)
            self.redGain.setSliderPosition(158)
            self.redGainBox.setValue(158)
        elif awb == 6:
            mode = "tungsten"
            self.sendCtrl(AWB_MODE + "off")
            self.sendCtrl(GAIN_BLUE + str(186))
            self.sendCtrl(GAIN_RED + str(114))
            self.blueGain.setSliderPosition(186)
            self.blueGainBox.setValue(186)
            self.redGain.setSliderPosition(114)
            self.redGainBox.setValue(114)
        elif awb == 7:
            mode = "fluorescent"
            self.sendCtrl(AWB_MODE + "off")
            self.sendCtrl(GAIN_BLUE + str(157))
            self.sendCtrl(GAIN_RED + str(127))
            self.blueGain.setSliderPosition(157)
            self.blueGainBox.setValue(157)
            self.redGain.setSliderPosition(127)
            self.redGainBox.setValue(127)
        elif awb == 8:
            mode = "incandescent"
            self.sendCtrl(AWB_MODE + "off")
            self.sendCtrl(GAIN_BLUE + str(195))
            self.sendCtrl(GAIN_RED + str(110))
            self.blueGain.setSliderPosition(195)
            self.blueGainBox.setValue(195)
            self.redGain.setSliderPosition(110)
            self.redGainBox.setValue(110)
        elif awb == 9:
            mode = "flash"
            self.sendCtrl(AWB_MODE + "off")
            self.sendCtrl(GAIN_BLUE + str(155))
            self.sendCtrl(GAIN_RED + str(128))
            self.blueGain.setSliderPosition(155)
            self.blueGainBox.setValue(155)
            self.redGain.setSliderPosition(128)
            self.redGainBox.setValue(128)
        else:
            return
        self.setColorsEnabled(not awb)
        self.redGain.blockSignals(False)
        self.redGainBox.blockSignals(False)
        self.blueGain.blockSignals(False)
        self.blueGainBox.blockSignals(False)
        self.updateStatus("Set white balance")

    # White balance for HQ camera.
    # awb_mode = "off" allows the camera's white balance to be set manually, by
    # adjusting the awb_gains parameter.
    def setAWBmodeHQ(self, awb):
        self.redGain.blockSignals(True)
        self.redGainBox.blockSignals(True)
        self.blueGain.blockSignals(True)
        self.blueGainBox.blockSignals(True)
        if awb == 0:
            mode = "off"
            self.blueGain.setSliderPosition(self.manualBlueGain)
            self.blueGainBox.setValue(self.manualBlueGain)
            self.redGain.setSliderPosition(self.manualRedGain)
            self.redGainBox.setValue(self.manualRedGain)
            self.sendCtrl(AWB_MODE + mode)
            self.sendCtrl(GAIN_BLUE + str(self.blueGainBox.value()))
            self.sendCtrl(GAIN_RED + str(self.redGainBox.value()))
        # Gain values for 3300 K color temperature.
        elif awb == 1:
            self.sendCtrl(AWB_MODE + "off")
            self.sendCtrl(GAIN_BLUE + str(232))
            self.sendCtrl(GAIN_RED + str(253))
            self.blueGain.setSliderPosition(232)
            self.blueGainBox.setValue(232)
            self.redGain.setSliderPosition(253)
            self.redGainBox.setValue(253)
        elif awb == 2:
            mode = "auto"
            self.sendCtrl(AWB_MODE + mode)
        elif awb == 3:
            mode = "sunlight"
            self.sendCtrl(AWB_MODE + mode)
            self.blueGain.setSliderPosition(151)
            self.blueGainBox.setValue(151)
            self.redGain.setSliderPosition(324)
            self.redGainBox.setValue(324)
        elif awb == 4:
            mode = "cloudy"
            self.sendCtrl(AWB_MODE + mode)
            self.blueGain.setSliderPosition(126)
            self.blueGainBox.setValue(126)
            self.redGain.setSliderPosition(372)
            self.redGainBox.setValue(372)
        elif awb == 5:
            mode = "shade"
            self.sendCtrl(AWB_MODE + mode)
            self.blueGain.setSliderPosition(126)
            self.blueGainBox.setValue(126)
            self.redGain.setSliderPosition(372)
            self.redGainBox.setValue(372)
        elif awb == 6:
            mode = "tungsten"
            self.sendCtrl(AWB_MODE + mode)
            self.blueGain.setSliderPosition(229)
            self.blueGainBox.setValue(229)
            self.redGain.setSliderPosition(247)
            self.redGainBox.setValue(247)
        elif awb == 7:
            mode = "fluorescent"
            self.sendCtrl(AWB_MODE + mode)
            self.blueGain.setSliderPosition(209)
            self.blueGainBox.setValue(209)
            self.redGain.setSliderPosition(260)
            self.redGainBox.setValue(260)
        elif awb == 8:
            mode = "incandescent"
            self.sendCtrl(AWB_MODE + mode)
            self.blueGain.setSliderPosition(253)
            self.blueGainBox.setValue(253)
            self.redGain.setSliderPosition(224)
            self.redGainBox.setValue(224)
        elif awb == 9:
            mode = "flash"
            self.sendCtrl(AWB_MODE + mode)
            self.blueGain.setSliderPosition(209)
            self.blueGainBox.setValue(209)
            self.redGain.setSliderPosition(260)
            self.redGainBox.setValue(260)
        else:
            return
        self.setColorsEnabled(not awb)
        self.redGain.blockSignals(False)
        self.redGainBox.blockSignals(False)
        self.blueGain.blockSignals(False)
        self.blueGainBox.blockSignals(False)
        self.updateStatus("Set white balance")

    def setColorsEnabled(self, enable):
        self.redGain.setEnabled(enable)
        self.redGainBox.setEnabled(enable)
        self.redResetBtn.setEnabled(enable)
        self.blueGain.setEnabled(enable)
        self.blueGainBox.setEnabled(enable)
        self.blueResetBtn.setEnabled(enable)

    # colorSetBtn
    def colorSet(self):
        self.awbBox.setCurrentIndex(0)
        self.updateStatus("Manual white balance")

    # blueGainBox
    def setGainBlue(self, value):
        if self.awbBox.currentIndex() == 0:
            self.manualBlueGain = value
            self.sendCtrl(GAIN_BLUE + str(value))
            self.updateStatus("Blue gain manual adjustment")

    # blueResetBtn
    def resetGainBlue(self):
        if config.cameraType == "V1":
            self.blueGainBox.setValue(100)
            
        elif config.cameraType == "HQ":        
            self.blueGainBox.setValue(160)
        
        self.updateStatus("Blue gain setting reset")

    # redGainBox
    def setGainRed(self, value):
        if self.awbBox.currentIndex() == 0:
            self.manualRedGain = value
            self.sendCtrl(GAIN_RED + str(value))
            self.updateStatus("Red gain manual adjustment")

    # redResetBtn
    def resetGainRed(self):
        if config.cameraType == "V1":
            self.redGainBox.setValue(100)
            
        elif config.cameraType == "HQ":
            self.redGainBox.setValue(320)
        
        self.updateStatus("Red gain setting reset")

    # brightnessBox
    def setBright(self, brt):
        self.sendCtrl(BRIGHTNESS + str(brt))
        self.updateStatus("Brightness level set")

    # resetBrightBtn
    def resetBright(self):
        self.brightnessBox.setValue(50)
        self.updateStatus("Brightness level reset")

    # contrastBox
    def setContrast(self, con):
        self.sendCtrl(CONTRAST + str(con))
        self.updateStatus("Contrast level set")

    # resetContBtn
    def resetContrast(self):
        self.contrastBox.setValue(0)
        self.updateStatus("Contrast level reset")

    # saturationBox
    def setSaturation(self, sat):
        self.sendCtrl(SATURATION + str(sat))
        self.updateStatus("Saturation level set")

    # resetSaturationBtn
    def resetSaturation(self):
        self.saturationBox.setValue(0)
        self.updateStatus("Saturation level reset")

    # quitButton
    def quitApp(self):
        msgBox = QMessageBox()
        msgBox.setWindowTitle("Application exit")
        msgBox.setIcon(QMessageBox.Question)
        msgBox.setText(" " * 5 + "Exit application?" + " " * 25)
        acceptButton = msgBox.addButton("Accept", QMessageBox.AcceptRole)
        rejectButton = msgBox.addButton("Cancel", QMessageBox.RejectRole)

        msgBox.exec()
        if msgBox.clickedButton() == acceptButton:
            # The current configuration is saved.
            self.saveCurrentConfig()
            # We close the application.
            self.closeEvent()

        else:
            return

    # Capture.

    # endFrameBox
    def setEndFrame(self, frame):
        config.frame_limit = frame
        self.updateStatus("Last frame to digitize: " +
                          str(config.frame_limit))

    # bracketingBox
    def setBlend(self, bracketing):
        self.sendCtrl(BRACKETING_SHOTS + str(bracketing))        

        if bracketing == 1:
            self.stopsBox.setEnabled(False)
            self.DS8tabWidget.setTabEnabled(4, False)
            self.setHDRAlgorithm(False)
            self.setToneMapAlgorithm(False)
            self.stopsBox.setValue(0.0)

        else:
            config.exposureTimes.resize(bracketing, refcheck=False)
            self.stopsBox.setEnabled(True)
            self.DS8tabWidget.setTabEnabled(4, True)
            self.setHDRAlgorithm(False)
            self.setToneMapAlgorithm(False)

        self.updateStatus("Set number of photos per frame")

    # captureTestBtn
    def takeTestPhoto(self):
        self.disableCaptureWidgets()
        config.lastMode = "T"
        config.prevOn = False
        config.testImg = True
        config.captureOn = False
        self.sendCtrl(TEST_PHOTO)
        self.updateStatus("Taken test image")

    # stopsBox
    def setBlendStops(self, stops):
        self.sendCtrl(BRACKETING_STOPS + str(stops))
        minExp = self.minExp(self.exposureBox.value())
        self.exposureBoxMin.setValue(minExp)
        maxExp = self.maxExp(self.exposureBox.value())
        self.exposureBoxMax.setValue(maxExp)
        self.updateStatus("Set number of stop points")

    # autoExpCheckBox
    def setAutoExp(self, isOn):
        self.sendCtrl(AUTOEXP_ON if isOn else AUTOEXP_OFF
                      + str(self.timeExpBox.value()))
        self.timeExpBox.blockSignals(isOn)
        self.timeExpBox.setEnabled(not isOn)

        if self.timeExpBox.value() and (not isOn):
            self.exposureDownBtn.setEnabled(True)
        else:
            self.exposureDownBtn.setEnabled(False)

        if (self.timeExpBox.value() < 1000.0) and (not isOn):
            self.exposureUpBtn.setEnabled(True)
        else:
            self.exposureUpBtn.setEnabled(False)

        if not isOn:
            # Manually set exposure time is saved for later use.
            texp = self.timeExpBox.value()
            self.manExpTime = texp

            # Multiple exposure data is updated.
            minExp = self.minExp(texp)
            self.exposureBoxMin.setValue(minExp)
            self.exposureBox.setValue(texp)
            maxExp = self.maxExp(texp)
            self.exposureBoxMax.setValue(maxExp)

            self.updateStatus("Set manual exposure")

        else:
            self.updateStatus("Set auto exposure")

    # timeExpBox
    def setExposure(self, texp):
        if not self.autoExpCheckBox.isChecked():
            texp = round(texp, 1)
            # Manually set exposure time is saved for later use.
            self.manExpTime = texp

            self.sendCtrl(FIX_EXPOSURE + str(texp))

            # Multiple exposure data is updated.
            minExp = self.minExp(texp)
            self.exposureBoxMin.setValue(minExp)
            self.exposureBox.setValue(texp)
            maxExp = self.maxExp(texp)
            self.exposureBoxMax.setValue(maxExp)

            if self.timeExpBox.value():
                self.exposureDownBtn.setEnabled(True)
            else:
                self.exposureDownBtn.setEnabled(False)

            if self.timeExpBox.value() < 1000.0:
                self.exposureUpBtn.setEnabled(True)
            else:
                self.exposureUpBtn.setEnabled(False)

            self.updateStatus("Set manual exposure")

    # exposureDownBtn
    def setExposureDown(self):
        texp = self.timeExpBox.value()
        texp *= 0.8
        texp = round(texp, 1)
        self.sendCtrl(FIX_EXPOSURE + str(texp))
        self.timeExpBox.setValue(texp)
        if self.timeExpBox.value():
            self.exposureDownBtn.setEnabled(True)
        if self.timeExpBox.value() < 1000.0:
            self.exposureUpBtn.setEnabled(True)
        self.updateStatus("Decreased exposure time 20%")

    # exposureUpBtn
    def setExposureUp(self):
        texp = self.timeExpBox.value()
        texp *= 1.2
        texp = round(texp, 1)
        if texp > 1000.0:
            texp = 1000.0
        self.sendCtrl(FIX_EXPOSURE + str(texp))
        self.timeExpBox.setValue(texp)
        if self.timeExpBox.value():
            self.exposureDownBtn.setEnabled(True)
        if self.timeExpBox.value() >= 1000.0:
            self.exposureUpBtn.setEnabled(False)
        self.updateStatus("Increased exposure time 20%")

    # captureFolderBox
    def checkCaptureFolder(self, capFolder):        
        
        capFolder = capFolder.strip()
        
        if capFolder and Path(capFolder).is_dir():
            
            config.capFolder = capFolder
            info("Capture folder: " + config.capFolder)
            self.updateStatus("Capture folder selected")
            if self.startPosCheckBox.isChecked():
                self.captureTestBtn.setEnabled(True)
                self.captureStartBtn.setEnabled(True)
                
        else:
            self.capFolderBox.setText("Select capture folder")
            
    
    # chooseFolderBtn
    def chooseFolder(self):

        filtros = ["Directories"]
        fileDialog = QFileDialog()
        fileDialog.setOption(QFileDialog.ShowDirsOnly, True)
        fileDialog.setFileMode(QFileDialog.DirectoryOnly)        
        fileDialog.setWindowTitle("Select capture folder")
        fileDialog.setDirectory(config.default_folder)
        fileDialog.setNameFilters(filtros)

        if fileDialog.exec():
            
            self.capFolderBox.setText(fileDialog.selectedFiles()[0])

        else:
            return

    # activateMotorCheckBox
    def activateMotor(self, checked):
        if checked:
            self.sendCtrl(ACTIVATE_MOTOR)
            self.updateStatus("Motor on")

        else:
            self.sendCtrl(DEACTIVATE_MOTOR)
            self.updateStatus("Motor off")

    # frameLCDCountUpBtn
    def positionIndicatorUp(self):
        if self.startPosCheckBox.isChecked():
            config.frame_number += 1
            config.file_number = config.frame_number
            self.dirty = False
            self.frameLcd.display(config.frame_number)

    # frameLCDCountDownBtn
    def positionIndicatorDown(self):
        if self.startPosCheckBox.isChecked():
            config.frame_number -= 1
            config.file_number = config.frame_number
            self.dirty = False
            self.frameLcd.display(config.frame_number)

    # gotoCheckBox
    def gotoFrameNumber(self, checked):
        if checked:
            numFrames = self.nextFrameBox.value() - config.frame_number
            if numFrames > 0:
                # Motor is activated.
                if not self.activateMotorCheckBox.isChecked():
                    self.sendCtrl(ACTIVATE_MOTOR)
                    self.activateMotorCheckBox.setChecked(True)
                self.sendCtrl(CAP_FRAME_ADV + str(numFrames))
                self.dirty = True
                self.motorNotMoving = False
                self.updateStatus("Forward to the frame " +
                                  str(self.nextFrameBox.value()))
                self.disableButtons()
                self.gotoCheckbox.setEnabled(True)

            elif numFrames < 0:
                numFrames *= -1
                # Motor is activated.
                if not self.activateMotorCheckBox.isChecked():
                    self.sendCtrl(ACTIVATE_MOTOR)
                    self.activateMotorCheckBox.setChecked(True)
                self.sendCtrl(CAP_FRAME_REV + str(numFrames))
                self.dirty = True
                self.motorNotMoving = False
                self.updateStatus("Backward to the frame " +
                                  str(self.nextFrameBox.value()))
                self.disableButtons()
                self.gotoCheckbox.setEnabled(True)

            elif numFrames == 0:
                self.gotoCheckbox.setChecked(False)
                return

        else:

            if self.motorNotMoving:
                return
            else:
                self.sendCtrl(MOTOR_STOP)
                self.updateStatus("Forward/Backward to the frame " +
                                  str(self.nextFrameBox.value()) +
                                  " cancelled")

    # passToCheckbox
    def passToLCDIndicator(self):
        frame = self.nextFrameBox.value()
        self.frameLcd.display(frame)
        config.frame_number = frame
        config.file_number = frame
        self.passToCheckbox.setChecked(False)

    # nextFrameBox does not require slot.

    # captureFrmRev10
    def capFrameRev10(self):
        # Motor is activated.
        if not self.activateMotorCheckBox.isChecked():
            self.sendCtrl(ACTIVATE_MOTOR)
            self.activateMotorCheckBox.setChecked(True)
        self.sendCtrl(CAP_FRAME_REV + "10")
        self.dirty = True
        self.motorNotMoving = False
        self.updateStatus("10 frame reverse")
        self.disableButtons()

    # captureFrmRev
    def capFrameRev(self):
        # Motor is activated.
        if not self.activateMotorCheckBox.isChecked():
            self.sendCtrl(ACTIVATE_MOTOR)
            self.activateMotorCheckBox.setChecked(True)
        self.sendCtrl(CAP_FRAME_REV)
        self.dirty = True
        self.motorNotMoving = False
        self.updateStatus("1 frame reverse")
        self.disableButtons()

    # captureFrmAdv
    def capFrameAdv(self):
        # Motor is activated.
        if not self.activateMotorCheckBox.isChecked():
            self.sendCtrl(ACTIVATE_MOTOR)
            self.activateMotorCheckBox.setChecked(True)
        self.sendCtrl(CAP_FRAME_ADV)
        self.dirty = True
        self.motorNotMoving = False
        self.updateStatus("1 frame advance")
        self.disableButtons()

    # captureFrmAdv10
    def capFrameAdv10(self):
        # Motor is activated.
        if not self.activateMotorCheckBox.isChecked():
            self.sendCtrl(ACTIVATE_MOTOR)
            self.activateMotorCheckBox.setChecked(True)
        self.sendCtrl(CAP_FRAME_ADV + "10")
        self.dirty = True
        self.motorNotMoving = False
        self.updateStatus("10 frame advance")
        self.disableButtons()

    # captureStopBtn
    @ pyqtSlot()
    def captureEnd(self):
        config.captureOn = False
        self.sendCtrl(STOP_CAPTURE)
        self.updateStatus("Capture stopped")
        self.enableCaptureWidgets()

    # captureStartBtn
    def captureStart(self):
        numImgFiles = len(glob(config.capFolder.strip() + "/img*.jpg"))
        if numImgFiles:
            msgBox = QMessageBox()
            msgBox.setWindowTitle("Existing image files")
            msgBox.setIcon(QMessageBox.Question)
            msgBox.setText("The selected folder contains " +
                           str(numImgFiles) + " image files.")
            msgBox.setInformativeText("Overwrite existing files?")
            acceptButton = msgBox.addButton("Accept", QMessageBox.AcceptRole)
            rejectButton = msgBox.addButton("Cancel", QMessageBox.RejectRole)

            msgBox.exec()
            if msgBox.clickedButton() == rejectButton:
                return
        if (config.imgCapResW % 2):
            msgBox = QMessageBox()
            msgBox.setWindowTitle("Image dimensions error")
            msgBox.setIcon(QMessageBox.Warning)
            msgBox.setText("The width of the resulting image (" +
                           str(config.imgCapResW) +
                           " px) is not divisible by 2.")
            msgBox.setInformativeText("Width modification required.")
            acceptButton = msgBox.addButton("Accept", QMessageBox.AcceptRole)

            msgBox.exec()
            if msgBox.clickedButton() == acceptButton:
                return

        # Motor is activated.
        if not self.activateMotorCheckBox.isChecked():
            self.sendCtrl(ACTIVATE_MOTOR)
            self.activateMotorCheckBox.setChecked(True)

        # Disable controls that affect capture.
        self.disableCaptureWidgets()

        # Start capture mode.
        config.lastMode = "C"
        config.prevOn = False
        config.testImg = False
        config.captureOn = True
        self.sendCtrl(START_CAPTURE)
        info("Waiting for image " + str(config.numImgRec))
        self.updateStatus("Capture mode started")

    # capturePauseBtn
    def capturePause(self, pressed):
        if pressed:

            self.sendCtrl(STOP_CAPTURE)
            self.captureStopBtn.setEnabled(False)
            self.updateStatus("Capture paused")

        else:

            # Restart capture mode.
            self.sendCtrl(START_CAPTURE)
            info("Waiting for image " + str(config.numImgRec))
            self.captureStopBtn.setEnabled(True)
            self.updateStatus("Capture mode started")

    # Advanced settings.

    # vflip
    def setVFlip(self, isOn):
        self.sendCtrl(VFLIP_ON if isOn else VFLIP_OFF)
        self.updateStatus("Set vertical flip")

    # hflip
    def setHFlip(self, isOn):
        self.sendCtrl(HFLIP_ON if isOn else HFLIP_OFF)
        self.updateStatus("Set horizontal flip")

    # bw
    def setBW(self, isOn):
        self.sendCtrl(BW_ON if isOn else BW_OFF)
        if isOn:
            self.updateStatus("B/W image")
        else:
            self.updateStatus("Color image")

    # drcBox
    def setDrc(self, drc):
        if drc == 0:
            self.sendCtrl(DRC + "off")
        elif drc == 1:
            self.sendCtrl(DRC + "low")
        elif drc == 2:
            self.sendCtrl(DRC + "medium")
        elif drc == 3:
            self.sendCtrl(DRC + "high")
        else:
            return
        self.updateStatus("Established DRC level")

    # sharpnessBox
    def setSharpness(self, sharp):
        self.sendCtrl(SHARPNESS + str(sharp))
        self.updateStatus("Set level of sharpness")

    # resolutionBox
    def setCapRes(self, idx):

        self.sendCtrl(SET_SIZE + str(idx))

        # The resolution of the captured images is updated.
        if config.cameraType == "V1":

            if idx == 0:
                self.imgCapW = 2592
                self.imgCapH = 1944
            elif idx == 1:
                self.imgCapW = 1600
                self.imgCapH = 1200
            elif idx == 2:
                self.imgCapW = 1440
                self.imgCapH = 1080
            elif idx == 3:
                self.imgCapW = 1296
                self.imgCapH = 972
            elif idx == 4:
                self.imgCapW = 1024
                self.imgCapH = 768
            elif idx == 5:
                self.imgCapW = 800
                self.imgCapH = 600

        elif config.cameraType == "HQ":

            if idx == 0:
                self.imgCapW = 2028
                self.imgCapH = 1080
            elif idx == 1:
                self.imgCapW = 2028
                self.imgCapH = 1520
            elif idx == 2:
                self.imgCapW = 4056
                self.imgCapH = 3040

        # Capture size and maximum allowed cropping are updated.
        self.setMaxCrop()

        self.updateStatus("Set capture resolution")

    # stillCap
    def setStillPort(self, still):
        self.sendCtrl(VID_PORT_OFF if still else VID_PORT_ON)
        if still:
            self.updateStatus("Selected still image port")
        else:
            self.updateStatus("Selected video port")

    # Post-capture settings.
    # They are made on the client. No need to inform the server.

    # showHist
    def initHistogram(self, isOn):
        config.showHist = isOn
        if isOn:
            if not self.histograma:
                self.histograma = DS8Histogram()
            self.histograma.mngr.window.setGeometry(1000, 200, 480, 360)
            self.logarithmHist.setEnabled(True)
            self.plotHistogramEvent.wait(10)
            self.updateHistogram(self.lastShowImg, self.lastShowTitle)
            self.updateStatus("Show histogram")
        else:
            self.histograma.closeHist()
            self.histograma = None
            self.logarithmHist.setEnabled(False)
            self.updateStatus("Don't show histogram")

    # logarithmHist
    def mklogHist(self, isOn):
        config.logarithmHist = isOn
        self.plotHistogramEvent.wait(10)
        self.updateHistogram(self.lastShowImg, self.lastShowTitle)

    # roundCorns
    def setRoundCorns(self, isOn):
        config.roundcorns = isOn
        if isOn:
            self.updateStatus("Round angles")
        else:
            self.updateStatus("Don't round angles")

        if isOn and self.prevCheckBox.isChecked():
            self.roundCorns.setEnabled(True)
        else:
            self.roundCorns.setEnabled(False)

    # rotation
    def setRotation(self, isOn):
        config.rotation = isOn
        if isOn:
            self.rotationBox.setEnabled(True)
            config.rotationValue = round(self.rotationBox.value(), 2)
            self.updateStatus("Rotate image")
        else:
            self.rotationBox.setEnabled(False)
            self.updateStatus("Do not rotate image")

        if isOn and self.prevCheckBox.isChecked():
            self.rotationBox.setEnabled(True)
        else:
            self.rotationBox.setEnabled(False)

    # rotationBox
    def setRotationValue(self, val):
        config.rotationValue = round(val, 2)
        if val:
            self.updateStatus("Rotation angle set")
        else:
            self.updateStatus("Angle of rotation at 0º")

    # cropping
    def setCrop(self, isOn):
        config.cropping = isOn
        if isOn:
            config.CropT = self.cropTopBox.value()
            config.cropL = self.cropLeftBox.value()
            config.cropR = self.cropRightBox.value()
            config.cropB = self.cropBottomBox.value()
            self.updateStatus("Crop image")

        else:
            self.updateStatus("Do not crop image")

        if isOn and self.prevCheckBox.isChecked():
            self.cropTopBox.setEnabled(True)
            self.cropLeftBox.setEnabled(True)
            self.cropRightBox.setEnabled(True)
            self.cropBottomBox.setEnabled(True)
        else:
            self.cropTopBox.setEnabled(False)
            self.cropLeftBox.setEnabled(False)
            self.cropRightBox.setEnabled(False)
            self.cropBottomBox.setEnabled(False)

    # cropTopBox

    def setCropT(self, val):
        config.cropT = val
        self.updateStatus("Set top cutout")

    # cropLeftBox
    def setCropL(self, val):
        config.cropL = val
        self.updateStatus("Set left clipping")

    # cropRightBox
    def setCropR(self, val):
        config.cropR = val
        self.updateStatus("Set right clipping")

    # cropBottomBox
    def setCropB(self, val):
        config.cropB = val
        self.updateStatus("Set bottom cutout")

    # HDR settings

    # HDRMertensRadioButton, HDRDebevecRadioButton
    def setHDRAlgorithm(self, updStat=True):
        if self.bracketingBox.value == 1:
            self.HDRMertensRadioButton.setEnabled(False)
            self.HDRDebevecRadioButton.setEnabled(False)
        else:
            self.HDRMertensRadioButton.setEnabled(True)
            self.HDRDebevecRadioButton.setEnabled(True)

        if self.HDRMertensRadioButton.isChecked():
            config.blender = "Mertens"
            self.SimpleRadioButton.setEnabled(False)
            self.SimpleGammaLabel.setEnabled(False)
            self.SimpleGammaSpinBox.setEnabled(False)
            self.ReinhardRadioButton.setEnabled(False)
            self.ReinhardGammaLabel.setEnabled(False)
            self.ReinhardGammaSpinBox.setEnabled(False)
            self.ReinhardIntensityLabel.setEnabled(False)
            self.ReinhardIntensitySpinBox.setEnabled(False)
            self.ReinhardLightLabel.setEnabled(False)
            self.ReinhardLightSpinBox.setEnabled(False)
            self.ReinhardColorLabel.setEnabled(False)
            self.ReinhardColorSpinBox.setEnabled(False)
            self.DragoRadioButton.setEnabled(False)
            self.DragoGammaLabel.setEnabled(False)
            self.DragoGammaSpinBox.setEnabled(False)
            self.DragoSaturationLabel.setEnabled(False)
            self.DragoSaturationSpinBox.setEnabled(False)
            self.DragoBiasLabel.setEnabled(False)
            self.DragoBiasSpinBox.setEnabled(False)
            self.MantiukRadioButton.setEnabled(False)
            self.MantiukGammaLabel.setEnabled(False)
            self.MantiukGammaSpinBox.setEnabled(False)
            self.MantiukSaturationLabel.setEnabled(False)
            self.MantiukSaturationSpinBox.setEnabled(False)
            self.MantiukScaleLabel.setEnabled(False)
            self.MantiukScaleSpinBox.setEnabled(False)
            if updStat:
                self.updateStatus("Set HDR Mertens")

        elif (self.HDRDebevecRadioButton.isEnabled() and
              self.HDRDebevecRadioButton.isChecked()):
            config.blender = "Debevec"
            self.SimpleRadioButton.setEnabled(True)
            self.ReinhardRadioButton.setEnabled(True)
            self.DragoRadioButton.setEnabled(True)
            self.MantiukRadioButton.setEnabled(True)
            self.setToneMapAlgorithm(False)
            if updStat:
                self.updateStatus("Set HDR Debevec")

    # Tone mapping algorithm controls.
    # SimpleRadioButton, ReinhardRadioButton, DragoRadioButton,
    # MantiukRadioButton
    def setToneMapAlgorithm(self, updStat=True):
        if (self.SimpleRadioButton.isEnabled() and
            self.SimpleRadioButton.isChecked()):
            config.toneMap = "Simple"
            self.SimpleGammaLabel.setEnabled(True)
            self.SimpleGammaSpinBox.setEnabled(True)
            self.ReinhardGammaLabel.setEnabled(False)
            self.ReinhardGammaSpinBox.setEnabled(False)
            self.ReinhardIntensityLabel.setEnabled(False)
            self.ReinhardIntensitySpinBox.setEnabled(False)
            self.ReinhardLightLabel.setEnabled(False)
            self.ReinhardLightSpinBox.setEnabled(False)
            self.ReinhardColorLabel.setEnabled(False)
            self.ReinhardColorSpinBox.setEnabled(False)
            self.DragoGammaLabel.setEnabled(False)
            self.DragoGammaSpinBox.setEnabled(False)
            self.DragoSaturationLabel.setEnabled(False)
            self.DragoSaturationSpinBox.setEnabled(False)
            self.DragoBiasLabel.setEnabled(False)
            self.DragoBiasSpinBox.setEnabled(False)
            self.MantiukGammaLabel.setEnabled(False)
            self.MantiukGammaSpinBox.setEnabled(False)
            self.MantiukSaturationLabel.setEnabled(False)
            self.MantiukSaturationSpinBox.setEnabled(False)
            self.MantiukScaleLabel.setEnabled(False)
            self.MantiukScaleSpinBox.setEnabled(False)
            if updStat:
                self.updateStatus("Established simple tone mapping algorithm")

        elif (self.ReinhardRadioButton.isEnabled() and
              self.ReinhardRadioButton.isChecked()):
            config.toneMap = "Reinhard"
            self.SimpleGammaLabel.setEnabled(False)
            self.SimpleGammaSpinBox.setEnabled(False)
            self.ReinhardGammaLabel.setEnabled(True)
            self.ReinhardGammaSpinBox.setEnabled(True)
            self.ReinhardIntensityLabel.setEnabled(True)
            self.ReinhardIntensitySpinBox.setEnabled(True)
            self.ReinhardLightLabel.setEnabled(True)
            self.ReinhardLightSpinBox.setEnabled(True)
            self.ReinhardColorLabel.setEnabled(True)
            self.ReinhardColorSpinBox.setEnabled(True)
            self.DragoGammaLabel.setEnabled(False)
            self.DragoGammaSpinBox.setEnabled(False)
            self.DragoSaturationLabel.setEnabled(False)
            self.DragoSaturationSpinBox.setEnabled(False)
            self.DragoBiasLabel.setEnabled(False)
            self.DragoBiasSpinBox.setEnabled(False)
            self.MantiukGammaLabel.setEnabled(False)
            self.MantiukGammaSpinBox.setEnabled(False)
            self.MantiukSaturationLabel.setEnabled(False)
            self.MantiukSaturationSpinBox.setEnabled(False)
            self.MantiukScaleLabel.setEnabled(False)
            self.MantiukScaleSpinBox.setEnabled(False)
            if updStat:
                self.updateStatus("Established Reinhard tone mapping algorithm")

        elif (self.DragoRadioButton.isEnabled() and
              self.DragoRadioButton.isChecked()):
            config.toneMap = "Drago"
            self.SimpleGammaLabel.setEnabled(False)
            self.SimpleGammaSpinBox.setEnabled(False)
            self.ReinhardGammaLabel.setEnabled(False)
            self.ReinhardGammaSpinBox.setEnabled(False)
            self.ReinhardIntensityLabel.setEnabled(False)
            self.ReinhardIntensitySpinBox.setEnabled(False)
            self.ReinhardLightLabel.setEnabled(False)
            self.ReinhardLightSpinBox.setEnabled(False)
            self.ReinhardColorLabel.setEnabled(False)
            self.ReinhardColorSpinBox.setEnabled(False)
            self.DragoGammaLabel.setEnabled(True)
            self.DragoGammaSpinBox.setEnabled(True)
            self.DragoSaturationLabel.setEnabled(True)
            self.DragoSaturationSpinBox.setEnabled(True)
            self.DragoBiasLabel.setEnabled(True)
            self.DragoBiasSpinBox.setEnabled(True)
            self.MantiukGammaLabel.setEnabled(False)
            self.MantiukGammaSpinBox.setEnabled(False)
            self.MantiukSaturationLabel.setEnabled(False)
            self.MantiukSaturationSpinBox.setEnabled(False)
            self.MantiukScaleLabel.setEnabled(False)
            self.MantiukScaleSpinBox.setEnabled(False)
            if updStat:
                self.updateStatus("Established Drago tone mapping algorithm")                                  

        elif (self.MantiukRadioButton.isEnabled() and
              self.MantiukRadioButton.isChecked()):
            config.toneMap = "Mantiuk"
            self.SimpleGammaLabel.setEnabled(False)
            self.SimpleGammaSpinBox.setEnabled(False)
            self.ReinhardGammaLabel.setEnabled(False)
            self.ReinhardGammaSpinBox.setEnabled(False)
            self.ReinhardIntensityLabel.setEnabled(False)
            self.ReinhardIntensitySpinBox.setEnabled(False)
            self.ReinhardLightLabel.setEnabled(False)
            self.ReinhardLightSpinBox.setEnabled(False)
            self.ReinhardColorLabel.setEnabled(False)
            self.ReinhardColorSpinBox.setEnabled(False)
            self.DragoGammaLabel.setEnabled(False)
            self.DragoGammaSpinBox.setEnabled(False)
            self.DragoSaturationLabel.setEnabled(False)
            self.DragoSaturationSpinBox.setEnabled(False)
            self.DragoBiasLabel.setEnabled(False)
            self.DragoBiasSpinBox.setEnabled(False)
            self.MantiukGammaLabel.setEnabled(True)
            self.MantiukGammaSpinBox.setEnabled(True)
            self.MantiukSaturationLabel.setEnabled(True)
            self.MantiukSaturationSpinBox.setEnabled(True)
            self.MantiukScaleLabel.setEnabled(True)
            self.MantiukScaleSpinBox.setEnabled(True)
            if updStat:
                self.updateStatus("Established Mantiuk tone mapping algorithm")                                  

    # SimpleGammaSpinBox
    def simpleGamma(self, gamma):
        gamma = round(gamma, 1)
        config.SimpleGamma = gamma
        self.updateStatus("Set gamma parameter to " + str(gamma))

    # ReinhardGammaSpinBox
    def reinhardGamma(self, gamma):
        gamma = round(gamma, 1)
        config.ReinhardGamma = gamma
        self.updateStatus("Set gamma parameter to " + str(gamma))

    # ReinhardIntensitySpinBox
    def reinhardIntensity(self, intensity):
        intensity = round(intensity, 1)
        config.ReinhardIntensity = intensity
        self.updateStatus("Intensity parameter set to " + str(intensity))

    # ReinhardLightSpinBox
    def reinhardLight(self, light):
        light = round(light, 1)
        config.ReinhardLight = light
        self.updateStatus("Set light parameter to " + str(light))

    # ReinhardColorSpinBox
    def reinhardColor(self, color):
        color = round(color, 1)
        config.ReinhardColor = color
        self.updateStatus("Set color parameter to" + str(color))

    # DragoGammaSpinBox
    def dragoGamma(self, gamma):
        gamma = round(gamma, 1)
        config.DragoGamma = gamma
        self.updateStatus("Set gamma parameter to " + str(gamma))

    # DragoSaturationSpinBox
    def dragoSaturation(self, saturation):
        saturation = round(saturation, 1)
        config.DragoSaturation = saturation
        self.updateStatus("Set saturation parameter to " + str(saturation))

    # DragoBiasSpinBox
    def dragoBias(self, bias):
        bias = round(bias, 2)
        config.DragoBias = bias
        self.updateStatus("Set bias parameter to " + str(bias))

    # MantiukGammaSpinBox
    def mantiukGamma(self, gamma):
        gamma = round(gamma, 1)
        config.MantiukGamma = gamma
        self.updateStatus("Set gamma parameter to " + str(gamma))

    # MantiukSaturationSpinBox
    def mantiukSaturation(self, saturation):
        saturation = round(saturation, 1)
        config.MantiukSaturation = saturation
        self.updateStatus("Set saturation parameter to " + str(saturation))                          

    # MantiukScaleSpinBox
    def mantiukScale(self, scale):
        scale = gamma(scale, 2)
        config.MantiukScale = scale
        self.updateStatus("Set scale parameter to " + str(scale))

    def disableButtons(self):
        self.fRevButton.setEnabled(False)
        self.reverseButton.setEnabled(False)
        self.forwardButton.setEnabled(False)
        self.ffdButton.setEnabled(False)
        self.captureTestBtn.setEnabled(False)
        self.activateMotorCheckBox.setEnabled(False)
        self.gotoCheckbox.setEnabled(False)
        self.passToCheckbox.setEnabled(False)
        self.nextFrameBox.setEnabled(False)
        self.captureFrmRev10.setEnabled(False)
        self.captureFrmRev.setEnabled(False)
        self.captureFrmAdv.setEnabled(False)
        self.captureFrmAdv10.setEnabled(False)
        self.captureStartBtn.setEnabled(False)
        self.quitButton.setEnabled(False)

    def enableButtons(self):
        self.fRevButton.setEnabled(True)
        self.reverseButton.setEnabled(True)
        self.forwardButton.setEnabled(True)
        self.ffdButton.setEnabled(True)
        self.quitButton.setEnabled(True)
        self.activateMotorCheckBox.setEnabled(True)
        if self.startPosCheckBox.isChecked():
            self.passToCheckbox.setEnabled(True)
            self.gotoCheckbox.setEnabled(True)
            self.nextFrameBox.setEnabled(True)
            self.captureFrmRev10.setEnabled(True)
            self.captureFrmRev.setEnabled(True)
            self.captureFrmAdv.setEnabled(True)
            self.captureFrmAdv10.setEnabled(True)
            if (config.capFolder.strip()
                and Path(config.capFolder.strip()).is_dir()):
                self.captureStartBtn.setEnabled(True)                
                if not config.prevOn:
                    self.captureTestBtn.setEnabled(True)

    def disableCaptureWidgets(self):
        # Disable lighting control.
        self.lightCheckbox.setEnabled(False)

        # Disable preview images.
        self.prevCheckBox.setChecked(False)
        self.prevCheckBox.setEnabled(False)
        config.prevOn = False

        # Disable home position indicator.
        self.startPosCheckBox.setEnabled(False)

        # Disable ISO setting.
        self.ISOLabel.setEnabled(False)
        self.ISOcomboBox.setEnabled(False)

        # Disable EV control.
        self.EVLabel.setEnabled(False)
        self.EV.setEnabled(False)

        # Disable white balance adjustment.
        self.awbLabel.setEnabled(False)
        self.awbBox.setEnabled(False)
        self.colorSetBtn.setEnabled(False)

        # Disable blue and red sliders and controls.
        self.blueLabel.setEnabled(False)
        self.blueGain.setEnabled(False)
        self.blueGainBox.setEnabled(False)
        self.blueResetBtn.setEnabled(False)
        self.redLabel.setEnabled(False)
        self.redGain.setEnabled(False)
        self.redGainBox.setEnabled(False)
        self.redResetBtn.setEnabled(False)

        # Disable slider and brightness control.
        self.brightnessLabel.setEnabled(False)
        self.brightness.setEnabled(False)
        self.brightnessBox.setEnabled(False)
        self.resetBrightBtn.setEnabled(False)

        # Disable sliders and controls for contrast and saturation.
        self.contrastLabel.setEnabled(False)
        self.contrast.setEnabled(False)
        self.contrastBox.setEnabled(False)
        self.resetContBtn.setEnabled(False)
        self.saturationLabel.setEnabled(False)
        self.saturation.setEnabled(False)
        self.saturationBox.setEnabled(False)
        self.resetSaturationBtn.setEnabled(False)

        # Disable exit button.
        self.quitButton.setEnabled(False)

        # Disable auto exposure control.
        self.autoExpCheckBox.setEnabled(False)

        # Disable motor activation control.
        self.activateMotorCheckBox.setEnabled(False)

        # Disable manual position adjustment controls.
        self.frameLCDCountUpBtn.setEnabled(False)
        self.frameLCDCountDownBtn.setEnabled(False)
        self.passToCheckbox.setEnabled(False)

        # Disable motion and capture controls.
        self.fRevButton.setEnabled(False)
        self.reverseButton.setEnabled(False)
        self.stopButton.setEnabled(False)
        self.forwardButton.setEnabled(False)
        self.ffdButton.setEnabled(False)
        self.bracketingBox.setEnabled(False)
        self.bracketingLabel.setEnabled(False)
        self.captureTestBtn.setEnabled(False)
        self.stopsBox.setEnabled(False)
        self.stopsLabel.setEnabled(False)
        self.timeExpLabel.setEnabled(False)
        self.timeExpBox.setEnabled(False)
        self.exposureDownBtn.setEnabled(False)
        self.exposureUpBtn.setEnabled(False)
        self.gotoCheckbox.setEnabled(False)
        self.nextFrameBox.setEnabled(False)
        self.captureFrmRev10.setEnabled(False)
        self.captureFrmRev.setEnabled(False)
        self.captureFrmAdv.setEnabled(False)
        self.captureFrmAdv10.setEnabled(False)
        self.captureStopBtn.setEnabled(True)
        self.captureStartBtn.setEnabled(False)
        self.capturePauseBtn.setEnabled(True)

        # Disable folder selection.
        self.configFileBox.setEnabled(False)
        self.loadConfigButton.setEnabled(False)
        self.capFolderBox.setEnabled(False)
        self.chooseFolderBtn.setEnabled(False)

        # Disable advanced settings.
        self.vflip.setEnabled(False)
        self.hflip.setEnabled(False)
        self.bw.setEnabled(False)
        self.drcLabel.setEnabled(False)
        self.drcBox.setEnabled(False)
        self.sharpnessLabel.setEnabled(False)
        self.sharpnessBox.setEnabled(False)
        self.resolutionLabel.setEnabled(False)
        self.resolutionBox.setEnabled(False)
        self.stillCap.setEnabled(False)

        # Disable post-capture settings.
        self.roundCorns.setEnabled(False)

        # Disable HDR settings.
        self.HDRMertensRadioButton.setEnabled(False)
        self.HDRDebevecRadioButton.setEnabled(False)
        
        self.SimpleRadioButton.setEnabled(False)
        if self.SimpleRadioButton.isChecked():
            self.SimpleGammaLabel.setEnabled(False)
            self.SimpleGammaSpinBox.setEnabled(False)
            
        self.ReinhardRadioButton.setEnabled(False)
        if self.ReinhardRadioButton.isChecked():
            self.ReinhardGammaLabel.setEnabled(False)
            self.ReinhardGammaSpinBox.setEnabled(False)
            self.ReinhardIntensityLabel.setEnabled(False)
            self.ReinhardIntensitySpinBox.setEnabled(False)
            self.ReinhardLightLabel.setEnabled(False)
            self.ReinhardLightSpinBox.setEnabled(False)
            self.ReinhardColorLabel.setEnabled(False)
            self.ReinhardColorSpinBox.setEnabled(False)
            
        self.DragoRadioButton.setEnabled(False)
        if self.DragoRadioButton.isChecked():
            self.DragoGammaLabel.setEnabled(False)
            self.DragoGammaSpinBox.setEnabled(False)            
            self.DragoSaturationLabel.setEnabled(False)
            self.DragoSaturationSpinBox.setEnabled(False)
            self.DragoBiasLabel.setEnabled(False)
            self.DragoBiasSpinBox.setEnabled(False)
            
        self.MantiukRadioButton.setEnabled(False)        
        if self.MantiukRadioButton.isChecked():
            self.MantiukGammaSpinBox.setEnabled(False)
            self.MantiukGammaLabel.setEnabled(False)            
            self.MantiukSaturationLabel.setEnabled(False)
            self.MantiukSaturationSpinBox.setEnabled(False)
            self.MantiukScaleLabel.setEnabled(False)
            self.MantiukScaleSpinBox.setEnabled(False)

    def enableCaptureWidgets(self):
        if not config.captureOn:
            # Enable lighting control.
            self.lightCheckbox.setEnabled(True)

            # Allow preview images.
            self.prevCheckBox.setEnabled(True)

            # Enable home position indicator.
            self.startPosCheckBox.setEnabled(True)

            # Allow ISO adjustment.
            self.ISOLabel.setEnabled(True)
            self.ISOcomboBox.setEnabled(True)

            # Enable EV control.
            self.EVLabel.setEnabled(True)
            self.EV.setEnabled(True)

            # Allow white balance adjustment.
            self.awbLabel.setEnabled(True)
            self.awbBox.setEnabled(True)
            self.colorSetBtn.setEnabled(True)

            # Enable blue and red sliders and controls.
            if self.awbBox.currentIndex() == 0:
                self.blueLabel.setEnabled(True)
                self.blueGain.setEnabled(True)
                self.blueGainBox.setEnabled(True)
                self.blueResetBtn.setEnabled(True)
                self.redLabel.setEnabled(True)
                self.redGain.setEnabled(True)
                self.redGainBox.setEnabled(True)
                self.redResetBtn.setEnabled(True)

            # Enable slider and brightness control.
            self.brightnessLabel.setEnabled(True)
            self.brightness.setEnabled(True)
            self.brightnessBox.setEnabled(True)
            self.resetBrightBtn.setEnabled(True)

            # Enable sliders and controls for contrast and saturation.
            self.contrastLabel.setEnabled(True)
            self.contrast.setEnabled(True)
            self.contrastBox.setEnabled(True)
            self.resetContBtn.setEnabled(True)
            self.saturationLabel.setEnabled(True)
            self.saturation.setEnabled(True)
            self.saturationBox.setEnabled(True)
            self.resetSaturationBtn.setEnabled(True)

            # Enable exit button.
            self.quitButton.setEnabled(True)

            # Enable auto exposure control.
            self.autoExpCheckBox.setEnabled(True)

            # Activate exposure increase/decrease buttons.
            self.exposureDownBtn.setEnabled(True)
            self.exposureUpBtn.setEnabled(True)

            # Enable motor activation control.
            self.activateMotorCheckBox.setEnabled(True)

            # Enable manual position adjustment controls.
            self.frameLCDCountUpBtn.setEnabled(True)
            self.frameLCDCountDownBtn.setEnabled(True)
            self.passToCheckbox.setEnabled(True)

            # Enable motion and capture controls.
            self.fRevButton.setEnabled(True)
            self.reverseButton.setEnabled(True)
            self.stopButton.setEnabled(False)
            self.forwardButton.setEnabled(True)
            self.ffdButton.setEnabled(True)
            self.bracketingBox.setEnabled(True)
            self.bracketingLabel.setEnabled(True)
            if self.startPosCheckBox.isChecked():
                self.captureTestBtn.setEnabled(True)
                self.captureStartBtn.setEnabled(True)
            if self.bracketingBox.value() > 1:
                self.stopsBox.setEnabled(True)
                self.stopsLabel.setEnabled(True)
            if not self.autoExpCheckBox.isChecked():
                self.timeExpLabel.setEnabled(True)
                self.timeExpBox.setEnabled(True)
                self.timeExpBox.setValue(self.manExpTime)
            if self.timeExpBox.value():
                self.exposureDownBtn.setEnabled(True)
            if self.timeExpBox.value() < 1000.0:
                self.exposureUpBtn.setEnabled(True)
            else:
                self.exposureUpBtn.setEnabled(False)
            self.gotoCheckbox.setEnabled(True)
            self.nextFrameBox.setEnabled(True)
            self.captureFrmRev10.setEnabled(True)
            self.captureFrmRev.setEnabled(True)
            self.captureFrmAdv.setEnabled(True)
            self.captureFrmAdv10.setEnabled(True)
            self.captureStopBtn.setEnabled(False)
            self.capturePauseBtn.setEnabled(False)
            self.nextFrameBox.setEnabled(True)
            self.endFrameBox.setEnabled(True)

            # Allow folder selection.
            self.configFileBox.setEnabled(True)
            self.loadConfigButton.setEnabled(True)
            self.capFolderBox.setEnabled(True)
            self.chooseFolderBtn.setEnabled(True)

            # Enable advanced settings.
            self.resolutionLabel.setEnabled(True)
            self.resolutionBox.setEnabled(True)
            self.stillCap.setEnabled(True)

            # Enable HDR settings.
            if self.bracketingBox.value() > 1:
                self.HDRMertensRadioButton.setEnabled(True)
                self.HDRDebevecRadioButton.setEnabled(True)

                if self.HDRDebevecRadioButton.isChecked():
                    self.SimpleRadioButton.setEnabled(True)
                    if self.SimpleRadioButton.isChecked():
                        self.SimpleGammaLabel.setEnabled(True)
                        self.SimpleGammaSpinBox.setEnabled(True)
                    self.ReinhardRadioButton.setEnabled(True)
                    if self.ReinhardRadioButton.isChecked():
                        self.ReinhardGammaLabel.setEnabled(True)
                        self.ReinhardGammaSpinBox.setEnabled(True)
                        self.ReinhardIntensityLabel.setEnabled(True)
                        self.ReinhardIntensitySpinBox.setEnabled(True)
                        self.ReinhardLightLabel.setEnabled(True)
                        self.ReinhardLightSpinBox.setEnabled(True)
                        self.ReinhardColorLabel.setEnabled(True)
                        self.ReinhardColorSpinBox.setEnabled(True)
                    self.DragoRadioButton.setEnabled(True)
                    if self.DragoRadioButton.isChecked():
                        self.DragoGammaLabel.setEnabled(True)
                        self.DragoGammaSpinBox.setEnabled(True)            
                        self.DragoSaturationLabel.setEnabled(True)
                        self.DragoSaturationSpinBox.setEnabled(True)
                        self.DragoBiasLabel.setEnabled(True)
                        self.DragoBiasSpinBox.setEnabled(True)
                    self.MantiukRadioButton.setEnabled(True)
                    if self.MantiukRadioButton.isChecked():
                        self.MantiukGammaSpinBox.setEnabled(True)
                        self.MantiukGammaLabel.setEnabled(True)            
                        self.MantiukSaturationLabel.setEnabled(True)
                        self.MantiukSaturationSpinBox.setEnabled(True)
                        self.MantiukScaleLabel.setEnabled(True)
                        self.MantiukScaleSpinBox.setEnabled(True)

                    self.setToneMapAlgorithm(False)

    # It is used in histogram activation operations and to activate controls
    # disabled while taking a test image.
    @ pyqtSlot(ndarray, str)
    def takenImg(self, img, title):
        self.lastShowImg = img
        self.lastShowTitle = title
        self.showImgSizes()
        if config.testImg:
            self.enableCaptureWidgets()
            config.testImg = False
            config.calibrate = False

    # Maximum allowed clipping.
    def setMaxCrop(self):

        # Minimum image size 100x100 pixels.
        maxCropW = int((self.imgCapW - 100) / 2)
        maxCropH = int((self.imgCapH - 100) / 2)
        self.cropTopBox.setMaximum(maxCropH)
        self.cropLeftBox.setMaximum(maxCropW)
        self.cropRightBox.setMaximum(maxCropW)
        self.cropBottomBox.setMaximum(maxCropH)

    def showImgSizes(self):

        self.sizeImgCapLabel.setText("Captured: " + str(config.imgCapIniW)
                                     + "x" + str(config.imgCapIniH))
        self.sizeImgFinalLabel.setText("Final: " + str(config.imgCapResW)
                                       + "x" + str(config.imgCapResH))

    # This function runs when the server reports engine stopped.
    @ pyqtSlot()
    def motorStopped(self):
        info("Engine stopped signal received")
        if self.dirty and self.startPosCheckBox.isChecked():
            config.file_number = config.frame_number
            self.dirty = False
        self.motorNotMoving = True
        if not self.capturePauseBtn.isChecked() and not config.captureOn:
            self.enableButtons()

    # This function is executed when the server reports light on or off. It is
    # used to update the status of the lighting control.
    @ pyqtSlot(str)
    def updateLightCheckbox(self, flag):
        # Light on.
        if flag == "l":
            info("Received light on signal")
            self.lightCheckbox.blockSignals(True)
            self.lightCheckbox.setEnabled(False)
            self.lightCheckbox.setChecked(True)

        # Light off.
        elif flag == "L":
            info("Received light off signal")
            self.lightCheckbox.setChecked(False)
            if not self.capturePauseBtn.isChecked():
                self.lightCheckbox.setEnabled(True)
                self.lightCheckbox.blockSignals(False)

        else:
            return

    @ pyqtSlot(int, int)
    def updateGains(self, gblue, gred):
        #info("updateGains called")
        info("Camera reported color gains. Blue = " + str(gblue) +
             ", Red = " + str(gred))
        self.redGain.blockSignals(True)
        self.redGainBox.blockSignals(True)
        self.blueGain.blockSignals(True)
        self.blueGainBox.blockSignals(True)

        self.blueGain.setSliderPosition(gblue)
        self.blueGainBox.setValue(gblue)
        self.redGain.setSliderPosition(gred)
        self.redGainBox.setValue(gred)

        self.redGain.blockSignals(False)
        self.redGainBox.blockSignals(False)
        self.blueGain.blockSignals(False)
        self.blueGainBox.blockSignals(False)

    # Automatic exposure data update.
    @ pyqtSlot(int, float, float, float)
    def updateSSAE(self, ssAE, again, dgain, framerate):
        if self.autoExpCheckBox.isChecked():
            #info("updateSSAE called")
            ssAE = round(ssAE / 1000, 1)
            minExp = self.minExp(ssAE)
            self.exposureBoxMin.setValue(minExp)
            self.exposureBox.setValue(ssAE)
            maxExp = self.maxExp(ssAE)
            self.exposureBoxMax.setValue(maxExp)
            self.timeExpBox.blockSignals(True)
            self.timeExpBox.setValue(ssAE)
            self.FPSdoubleSpinBox.setValue(framerate)
            self.gainBoxA.setValue(again)
            self.gainBoxD.setValue(dgain)
            self.timeExpBox.blockSignals(False)

    # Exposure data update.
    @ pyqtSlot(int, float, float, float)
    def updateSS(self, ss, again, dgain, framerate):
        #info("updateSS called")
        ss = round((ss / 1000), 1)
        self.timeExpBox.blockSignals(True)
        self.timeExpBox.setValue(ss)
        self.FPSdoubleSpinBox.setValue(framerate)
        self.gainBoxA.setValue(again)
        self.gainBoxD.setValue(dgain)
        if not config.captureOn and not config.testImg:
            self.timeExpBox.setValue(self.manExpTime)
        self.timeExpBox.blockSignals(False)

    def minExp(self, ss):
        if self.bracketingBox.value() == 1:
            minExp = round(ss, 1)
        else:
            minExp = round(ss / (2**(self.stopsBox.value() / 2)), 1)

        return minExp

    def maxExp(self, ss):
        if self.bracketingBox.value() == 1:
            maxExp = round(ss, 1)
        else:
            maxExp = round(ss * (2**(self.stopsBox.value() / 2)), 1)

        return maxExp

    @ pyqtSlot(str)
    def updateFrameNum(self, flag):
        if self.startPosCheckBox.isChecked():
            # Advance of a frame.
            if flag == "c":
                info("Frame advance signal received")
                config.frame_number += 1
            # Back one frame.
            elif flag == "C":
                info("Frame reverse signal received")
                config.frame_number -= 1
            # The position indicator is updated.
            self.frameLcd.display(config.frame_number)

            # The command go to is deactivated.
            if config.frame_number == self.nextFrameBox.value():
                self.gotoCheckbox.setChecked(False)

    @ pyqtSlot(str)
    def updateStatus(self, status):
        self.statusBar.setText(status)

    @ pyqtSlot(ndarray, str)
    def updateHistogram(self, cvimg, title):
        if config.showHist and self.histograma:
            self.histograma.plotHistogram(cvimg, title)
            self.plotHistogramEvent.set()
        else:
            return

    def closeEvent(self, *args, **kwargs):
        app = QApplication.instance()
        app.closeAllWindows()
        config.exitFlag = True
        self.sendCtrl(MOTOR_STOP)
        self.sendCtrl(LIGHT_OFF)
        if config.prevOn:
            self.sendCtrl(PREVIEW_OFF)
            sleep(1)
        self.sendCtrl(CLIENT_QUIT)
        info("Program exit")
