# -*- coding: utf-8 -*-

"""
DSuper8 project based on Joe Herman's rpi-film-capture.

Software modified by Manuel Ángel.

User interface redesigned by Manuel Ángel.

DS8ImgThread.py: Class and support functions for reading and processing the
                 images.

Last version: July 2021.
"""

from cv2 import (IMREAD_COLOR, resize, createMergeMertens, createMergeDebevec,
                 createCalibrateDebevec, getRotationMatrix2D, warpAffine,
                 imwrite, IMWRITE_JPEG_QUALITY, IMWRITE_PNG_COMPRESSION,
                 imdecode, imread, getOptimalNewCameraMatrix, createTonemap,
                 createTonemapReinhard, createTonemapDrago,
                 createTonemapMantiuk)

from threading import Event

from struct import unpack, calcsize

from logging import info

from numpy import uint8, shape, clip, ndarray, fromstring, bitwise_and

from io import BytesIO

from pathlib import Path

from PyQt5.QtCore import QThread, pyqtSignal

# Our own modules.

import config

from codes import NEW_IMAGE

# Class and support functions for the reading and treatment of images.
class imgThread(QThread):

    updateFrameNumSig = pyqtSignal(str)
    updateAESig = pyqtSignal(int, float, float, float)
    updateSSSig = pyqtSignal(int, float, float, float)
    updateGainsSig = pyqtSignal(int, int)
    updateStatusSig = pyqtSignal(str)
    displayImgSig = pyqtSignal(ndarray, str)
    plotHistogramSig = pyqtSignal(ndarray, str)
    motorStoppedSig = pyqtSignal()
    lightSig = pyqtSignal(str)
    frameLimitReachedSig = pyqtSignal()
    displayImgEvent = Event()
    displayImgEvent.set()
    plotHistogramEvent = Event()
    plotHistogramEvent.set()

    def __init__(self, connection, app):
        QThread.__init__(self, parent=app)
        self.threadID = 1
        self.name = "ImgThread"
        self.conn = connection

        # Temporary storage of the image received from the server.
        self.image_stream = BytesIO()

        # Image size received from server.
        self.image_len = None

        # Group of images from multiple exposures to obtain an HDR image.
        self.imglist = []

        # Index used for the exposure time matrix.
        self.indexETM = 0

        # Indicator of the type of information received from the server.
        self.imgflag = ""

        # Exposure time reported by camera.
        self.exposureTime = 0

        # Generate angle round masks.
        self.roundcornTLImg = config.readImgFromQrc(":/roundcornTL.png")
        self.roundcornTRImg = config.readImgFromQrc(":/roundcornTR.png")
        self.roundcornBRImg = config.readImgFromQrc(":/roundcornBR.png")
        self.roundcornBLImg = config.readImgFromQrc(":/roundcornBL.png")

        # Test image file name.
        self.testfilename = ""

    def updateAE(self, ssAE, again, dgain, framerate):
        self.updateAESig.emit(ssAE, again, dgain, framerate)

    def updateSS(self, ss, again, dgain, framerate):
        self.updateSSSig.emit(ss, again, dgain, framerate)

    def updateGains(self, gblue, gred):
        self.updateGainsSig.emit(gblue, gred)

    def updateStatus(self, status):
        self.updateStatusSig.emit(status)

    def postProcess(self, img):

        h, w = img.shape[:2]

        # Image rotation if selected.
        if config.rotation:
            #info("Rotating image")
            img = self.imageRotation(img, w, h)

        # Cropping the image if selected.
        if config.cropping:
            #info("Cropping image")
            img = img[config.cropT:h - config.cropB,
                      config.cropL:w - config.cropR]

        return img

    def imageRotation(self, img, w, h):
        rotMtx = getRotationMatrix2D((w / 2, h / 2), config.rotationValue, 1)
        img = warpAffine(img, rotMtx, (w, h))

        #info("Rotated image")

        return img

    # Scaling the image to the maximum dimensions specified in the config.py
    # file.
    def imageResize(self, img):
        h, w = img.shape[:2]

        config.imgCapIniH = h
        config.imgCapIniW = w

        finalh = config.imgCapFinalH
        finalw = int((finalh / h) * w)
        if finalw > config.imgCapFinalW:
            finalw = config.imgCapFinalW
            finalh = int((finalw / w) * h)
        img = resize(img, (finalw, finalh))

        #info("Resized image")

        return img

    def roundCorners(self, img):
        h, w = img.shape[:2]

        img[0:50, 0:50] = bitwise_and(img[0:50, 0:50], self.roundcornTLImg)
        img[0:50, w - 50:w] = bitwise_and(img[0:50, w - 50:w],
                                          self.roundcornTRImg)
        img[h - 50:h, w - 50:w] = bitwise_and(img[h - 50:h, w - 50:w],
                                              self.roundcornBRImg)
        img[h - 50:h, 0:50] = bitwise_and(img[h - 50:h, 0:50],
                                          self.roundcornBLImg)
        return img

    def blendImgList(self):

        if config.blender == "Mertens":
            blender = createMergeMertens()
            img = blender.process(self.imglist)

        else:
            # In tests carried out, it has been found that, to obtain good
            # results with the Debevec algorithm, it is required to take enough
            # images, in a number around 16.

            # Get the response function of the camera (CRF).
            calibrateDebevec = createCalibrateDebevec()
            responseDebevec = calibrateDebevec.process(self.imglist,
                                                       config.exposureTimes)

            # Merge the images into a linear HDR image.
            blender = createMergeDebevec()
            hdrDebevec = blender.process(self.imglist, config.exposureTimes,
                                         responseDebevec)

            # Apply tone mapping.
            if config.toneMap == "Simple":
                img = self.toneMapSimple(hdrDebevec)
            elif config.toneMap == "Reinhard":
                img = self.toneMapReinhard(hdrDebevec)
            elif config.toneMap == "Drago":
                img = self.toneMapDrago(hdrDebevec)
            elif config.toneMap == "Mantiuk":
                img = self.toneMapMantiuk(hdrDebevec)

        img = clip(img * 255, 0, 255).astype('uint8')

        info("Images fusion done")

        return img

    # Apply simple tone mapping.
    def toneMapSimple(self, hdrDebevec):

        toneMap = createTonemap(gamma=config.SimpleGamma)
        img = toneMap.process(hdrDebevec)
        return img

    # Apply Reinhard method tone mapping.
    def toneMapReinhard(self, hdrDebevec):

        toneMap = createTonemapReinhard(config.ReinhardGamma,
                                        config.ReinhardIntensity,
                                        config.ReinhardLight,
                                        config.ReinhardColor)

        img = toneMap.process(hdrDebevec)
        return img

    # Apply Drago method tone mapping.
    def toneMapDrago(self, hdrDebevec):

        toneMap = createTonemapDrago(config.DragoGamma,
                                     config.DragoSaturation,
                                     config.DragoBias)
        img = toneMap.process(hdrDebevec)
        img *= 3
        return img

    # Apply Mantiuk method tone mapping.
    def toneMapMantiuk(self, hdrDebevec):

        toneMap = createTonemapMantiuk(config.MantiukGamma, config.MantiukScale,
                                       config.MantiukSaturation)
        img = toneMap.process(hdrDebevec)
        img = 3 * img
        return img

    def showImage(self, img, title=""):
        self.displayImgEvent.wait(10)
        self.displayImgSig.emit(img, title)
        self.displayImgEvent.clear()

    def showHist(self, img, title=""):
        self.plotHistogramEvent.wait(10)
        self.plotHistogramSig.emit(img, title)
        self.plotHistogramEvent.clear()

    def finalizeImage(self, img):

        imageName = ""

        if config.lastMode == "P":
            imageName = "Previous " + str(config.numImgRec)

        elif config.lastMode == "T":
            i = 1
            while True:
                imageName = "Test{:05d}.jpg".format(i)

                self.testfilename = Path(config.capFolder.strip() + "/"
                                         + imageName)
                if self.testfilename.exists():
                    i += 1
                    continue
                else:
                    break

        elif config.lastMode == "C":
            # PNG image format has been tested. It is very slow.
            # imageName = "img{:05d}.png".format(config.file_number)

            imageName = "img{:05d}.jpg".format(config.file_number)

        # Image resize.
        img = self.imageResize(img)

        # The image histogram is displayed.
        if config.showHist:
            self.showHist(img, imageName)

        # Rounding the image angles.
        if config.roundcorns:
            img = self.roundCorners(img)

        # The image is shown.
        self.showImage(img, imageName)

        return img

    def writeImgFile(self, img):

        if config.testImg:
            filename = self.testfilename

        else:
            # PNG image format has been tested. It is very slow.
            # filename = (Path(config.capFolder.strip()
            #             + "/img{:05d}.png".format(config.file_number)))
            filename = (Path(config.capFolder.strip()
                             + "/img{:05d}.jpg".format(config.file_number)))

        with open(filename, "w") as imfile:
            # imwrite(str(filename), img,
            #        [int(IMWRITE_PNG_COMPRESSION), 5])
            imwrite(str(filename), img,
                    [int(IMWRITE_JPEG_QUALITY), 97])

        if config.testImg:
            info("Test image saved in: " + str(filename))

        else:
            info("Captured image saved in: " + str(filename))

    # With this function the server is requested to capture and send a new
    # image.
    def newImage(self):
        config.ctrl_conn.write(NEW_IMAGE + "\n")
        config.ctrl_conn.flush()
        info("Image " + str(config.numImgRec) + " requested")

    # Processing functions of the information received from the server.
    def imgflag_e(self):
        ssAE = unpack("<l", self.conn.read(calcsize("<l")))[0]
        again = unpack("<f", self.conn.read(calcsize("<f")))[0]
        dgain = unpack("<f", self.conn.read(calcsize("<f")))[0]
        framerate = unpack("<f", self.conn.read(calcsize("<f")))[0]
        self.updateAE(ssAE, again, dgain, framerate)

    def imgFlag_f(self):
        ss = unpack("<l", self.conn.read(calcsize("<l")))[0]
        again = unpack("<f", self.conn.read(calcsize("<f")))[0]
        dgain = unpack("<f", self.conn.read(calcsize("<f")))[0]
        framerate = unpack("<f", self.conn.read(calcsize("<f")))[0]
        self.updateSS(ss, again, dgain, framerate)

    def imgFlag_g(self):
        gblue = unpack("<i", self.conn.read(calcsize("<i")))[0]
        gred = unpack("<i", self.conn.read(calcsize("<i")))[0]
        self.updateGains(gblue, gred)

    def imgFlag_lL(self):

        self.lightSig.emit(self.imgflag)

    def imgFlag_spab(self):
        self.exposureTime = unpack("<i", self.conn.read(calcsize("<i")))[0]
        self.image_len = unpack("<L", self.conn.read(calcsize("<L")))[0]
        if not self.image_len:
            info("Output signal (image size 0)"
                 " received from server")
            config.exitFlag = True

        # Save image data to temporary storage.
        self.image_stream.write(self.conn.read(self.image_len))
        self.image_stream.seek(0)

    def imgFlag_p(self):
        cvimg = imdecode(fromstring(self.image_stream.read(self.image_len),
                                    dtype=uint8), IMREAD_COLOR)

        if type(cvimg) != ndarray:
            return

        info("Preview image " + str(config.numImgRec) + " received" +
             " - " + str(self.image_len) + " bytes - Exp. time " + 
             str(self.exposureTime) + " us")

        config.numImgRec += 1

        # New image is requested from the server.
        if config.prevOn:
            self.newImage()

        cvimg = self.postProcess(cvimg)
        cvimg = self.finalizeImage(cvimg)

    def imgFlag_s(self):
        cvimg = imdecode(fromstring(self.image_stream.read(self.image_len),
                                    dtype=uint8), IMREAD_COLOR)

        info("Single image " + str(config.numImgRec) + " received"
             + " - " + str(self.image_len) + " bytes - Exp. time "
             + str(self.exposureTime) + " us")

        config.numImgRec += 1

        # New image is requested from the server.
        if (config.captureOn and config.file_number < config.frame_limit):
            self.newImage()

        cvimg = self.postProcess(cvimg)
        cvimg = self.finalizeImage(cvimg)
        self.writeImgFile(cvimg)

        if config.lastMode == "C":

            if config.file_number >= config.frame_limit:
                self.frameLimitReachedSig.emit()
            config.file_number += 1

    def imgFlag_a(self):
        cvimg = imdecode(fromstring(self.image_stream.read(self.image_len),
                                    dtype=uint8), IMREAD_COLOR)

        info("Bracketing image " + str(config.numImgRec) + " received"  
             + " - " + str(self.image_len) + " bytes - Exp. time "
             + str(self.exposureTime) + " us")

        cvimg = self.postProcess(cvimg)

        # The image is saved in a list to later perform the fusion of the
        # images.

        self.imglist.append(cvimg)

        # The HDR Debevec algorithm uses time in s.
        config.exposureTimes[self.indexETM] = float(self.exposureTime
                                                    * 1e-6)
        self.indexETM += 1

    def imgFlag_b(self):
        cvimg = imdecode(fromstring(self.image_stream.read(self.image_len),
                                    dtype=uint8), IMREAD_COLOR)

        info("Last bracketing image " + str(config.numImgRec) + " received"
             + " - " + str(self.image_len) + " bytes - Exp. time "
             + str(self.exposureTime) + " us")

        config.numImgRec += 1

        # New image is requested from the server.
        if config.captureOn and (config.file_number < config.frame_limit):
            self.newImage()

        cvimg = self.postProcess(cvimg)

        # The image is saved in a list to later perform the fusion of the
        # images.
        self.imglist.append(cvimg)
        config.exposureTimes[self.indexETM] = float(self.exposureTime *
                                                    1e-6)
        cvimg = self.blendImgList()
        cvimg = self.finalizeImage(cvimg)
        self.writeImgFile(cvimg)

        # Cleaning the list of images and exposure times.
        self.imglist = []
        self.indexETM = 0

        if config.lastMode == "C":

            if config.file_number >= config.frame_limit:
                self.frameLimitReachedSig.emit()
            config.file_number += 1

    # Imaging thread main loop.

    def run(self):
        info("Executing main function of image thread")

        try:
            while not config.exitFlag:

                # The flag of the information sent by the server is obtained.
                self.imgflag = self.conn.read(1)
                try:
                    self.imgflag = self.imgflag.decode()
                except:
                    continue

                # Flag e -> automatic exposure time.
                if self.imgflag == "e":
                    self.imgflag_e()

                # Flag f -> exposure time and analog and digital gains.
                elif self.imgflag == "f":
                    self.imgFlag_f()

                # Flag g -> red and blue gains.
                elif self.imgflag == "g":
                    self.imgFlag_g()

                # Flags l -> light on and L -> light off
                elif self.imgflag == "l" or self.imgflag == "L":
                    self.imgFlag_lL()

                # Flag m -> motor stopped.
                elif self.imgflag == "m":
                    self.motorStoppedSig.emit()

                # Flag c -> frame advance
                # Flag C -> reverse of a frame
                elif self.imgflag == "c":
                    self.updateFrameNumSig.emit("c")

                elif self.imgflag == "C":
                    self.updateFrameNumSig.emit("C")

                # Flag p -> preview image.
                # Flag s -> single capture image.
                # Flag a -> image from a series of images to merge.
                # Flag b -> last image of the serie.

                else:

                    if (self.imgflag == "p" or self.imgflag == "s" or
                            self.imgflag == "a" or self.imgflag == "b"):
                        self.imgFlag_spab()

                    # Preview image.
                    if self.imgflag == "p":
                        self.imgFlag_p()

                    # Digitized frame with a single image.
                    if self.imgflag == "s":
                        self.imgFlag_s()

                    # Digitized frame with multiple images merged.
                    # One of several bracketting images.
                    if self.imgflag == "a":
                        self.imgFlag_a()

                    # The last of several bracketting images.
                    if self.imgflag == "b":
                        self.imgFlag_b()

                # Cleaning the temporary storage of the image received from the
                # server.
                self.image_stream.seek(0)
                self.image_stream.truncate()

        finally:
            info("Closing the thread " + self.name)
            self.conn.close()


